from typing import List, Tuple

import torch

from relogic.logickit.dataflow import DataFlow, Example, Feature, MiniBatch
from relogic.logickit.utils import create_tensor
from transformers.tokenization_utils import PreTrainedTokenizer
from relogic.logickit.tokenizer import FasttextTokenizer



class TwinsExample(Example):
  """
  Args:
    text_a (str): sentence in source language
    text_b (str): sentence in target language
    label (str)
  """
  def __init__(self, guid: str, text_a: str, text_b: str, labels: str = None):
    super().__init__()
    self.guid = guid
    self.text_a = text_a
    self.text_b = text_b
    self.labels = labels
    self.padding_id = 0

  def process(self, tokenizers, *inputs, **kwargs):
    """Process the sentence pair."""

    for tokenizer in tokenizers.values():
      if isinstance(tokenizer, PreTrainedTokenizer):
        self.text_a_tokens = tokenizer.tokenize(self.text_a)
        self.text_b_tokens = tokenizer.tokenize(self.text_b)

        self.text_a_tokens = self.text_a_tokens[:510]
        self.text_b_tokens = self.text_b_tokens[:510]

        self.a_tokens = ["[CLS]"] + self.text_a_tokens + ["[SEP]"]
        self.a_segment_ids = [0] * len(self.a_tokens)

        self.b_tokens = ["[CLS]"] + self.text_b_tokens + ["[SEP]"]
        self.b_segment_ids = [0] * len(self.b_tokens)

        self.a_input_ids = tokenizer.convert_tokens_to_ids(self.a_tokens)
        self.b_input_ids = tokenizer.convert_tokens_to_ids(self.b_tokens)
        self.a_input_mask = [1] * len(self.a_input_ids)
        self.b_input_mask = [1] * len(self.b_input_ids)

        if self.labels is not None:
          label_mapping = kwargs.get("label_mapping")
          if isinstance(self.labels, str):
            # Single Label Classfication
            self.label_ids = label_mapping[self.labels]
          elif isinstance(self.labels, list):
            # Multi Label Classification
            label_size = len(label_mapping)
            self.label_ids = [0] * label_size
            for label in self.labels:
              self.label_ids[label_mapping[label]] = 1
        else:
          self.label_ids = None
      elif isinstance(tokenizer, FasttextTokenizer):
        self._a_text_tokens = tokenizer.tokenize(self.text_a)
        self._b_text_tokens = tokenizer.tokenize(self.text_b)

        self._a_input_token_ids = tokenizer.convert_tokens_to_ids(self._a_text_tokens)
        self._b_input_token_ids = tokenizer.convert_tokens_to_ids(self._b_text_tokens)
        self._a_input_token_mask = [True] * len(self._a_input_token_ids)
        self._b_input_token_mask = [True] * len(self._b_input_token_ids)

        if self.labels is not None:
          label_mapping = kwargs.get("label_mapping")
          if isinstance(self.labels, str):
            # Single Label Classfication
            self._label_ids = label_mapping[self.labels]
          elif isinstance(self.labels, list):
            # Multi Label Classification
            label_size = len(label_mapping)
            self._label_ids = [0] * label_size
            for label in self.labels:
              self._label_ids[label_mapping[label]] = 1
        else:
          self._label_ids = None



  @classmethod
  def from_structure(cls, structure):
    return cls(guid="",
               text_a=structure.text_a,
               text_b=structure.text_b)

  @classmethod
  def from_json(cls, example):
    return cls(guid="{}|{}".format(example.get("text_a_id", 0), example.get("text_b_id", 0)),
               text_a=example["text_a"],
               text_b=example["text_b"],
               labels=example.get("label", None))

  @property
  def len(self):
    return len(self.a_tokens) + len(self.b_tokens)

  @property
  def len_a(self):
    return len(self.a_tokens)

  @property
  def len_b(self):
    return len(self.b_tokens)

  @property
  def _len(self):
    return len(self._a_input_token_ids) + len(self._b_input_token_ids)

  @property
  def _len_a(self):
    return len(self._a_input_token_ids)

  @property
  def _len_b(self):
    return len(self._b_input_token_ids)


class TwinsFeature(Feature):
  """Parallel Feature

  """
  def __init__(self, *input, **kwargs):
    super().__init__()
    self.a_input_ids = kwargs.pop("a_input_ids")
    self.b_input_ids = kwargs.pop("b_input_ids")
    self.a_segment_ids = kwargs.pop("a_segment_ids")
    self.b_segment_ids = kwargs.pop("b_segment_ids")
    self.a_input_mask = kwargs.pop("a_input_mask")
    self.b_input_mask = kwargs.pop("b_input_mask")
    self.label_ids = kwargs.pop("label_ids")

    # Classical Feature
    self._a_input_token_ids = kwargs.pop("_a_input_token_ids")
    self._b_input_token_ids = kwargs.pop("_b_input_token_ids")
    self._a_token_length = kwargs.pop("_a_token_length")
    self._b_token_length = kwargs.pop("_b_token_length")
    self._label_ids = kwargs.pop("_label_ids")
    self._a_input_token_mask = kwargs.pop("_a_input_token_mask")
    self._b_input_token_mask = kwargs.pop("_b_input_token_mask")


class TwinsMiniBatch(MiniBatch):
  def __init__(self, *inputs, **kwargs):
    super().__init__(*inputs, **kwargs)

  def generate_input(self, device, use_label):
    """Generate tensors based on Parallel Feature"""
    # BERT based features
    inputs = {}
    inputs["task_name"] = self.task_name
    inputs["a_input_ids"] = create_tensor(self.input_features, "a_input_ids",
                                          torch.long, device)
    inputs["b_input_ids"] = create_tensor(self.input_features, "b_input_ids",
                                          torch.long, device)
    inputs["a_input_mask"] = create_tensor(self.input_features, "a_input_mask",
                                           torch.long, device)
    inputs["b_input_mask"] = create_tensor(self.input_features, "b_input_mask",
                                           torch.long, device)
    inputs["a_segment_ids"] = create_tensor(self.input_features, "a_segment_ids",
                                            torch.long, device)
    inputs["b_segment_ids"] = create_tensor(self.input_features, "b_segment_ids",
                                            torch.long, device)

    if use_label:
      label_ids = create_tensor(self.input_features, "label_ids",
                                torch.long, device)
      inputs["label_ids"] = label_ids
    else:
      inputs["label_ids"] = None


    inputs["extra_args"] = {}
    if self.config.tasks[self.task_name]["selected_non_final_layers"] is not None:
      inputs["extra_args"]["selected_non_final_layers"] = self.config.tasks[self.task_name]["selected_non_final_layers"]

    # Classical Features
    inputs["_a_input_token_ids"] = create_tensor(self.input_features, "_a_input_token_ids",
                                                 torch.long, device)
    inputs["_b_input_token_ids"] = create_tensor(self.input_features, "_b_input_token_ids",
                                                 torch.long, device)
    inputs["_a_token_length"] = create_tensor(self.input_features, "_a_token_length",
                                              torch.long, device)
    inputs["_b_token_length"] = create_tensor(self.input_features, "_b_token_length",
                                              torch.long, device)
    inputs["_a_input_token_mask"] = create_tensor(self.input_features, "_a_input_token_mask",
                                                  torch.long, device)
    inputs["_b_input_token_mask"] = create_tensor(self.input_features, "_b_input_token_mask",
                                                  torch.long, device)

    if use_label:
      _label_ids = create_tensor(self.input_features, "_label_ids",
                                 torch.long, device)
      inputs["_label_ids"] = _label_ids
    else:
      inputs["_label_ids"] = None
    return inputs

class TwinsDataFlow(DataFlow):
  """DataFlow implementation based on Parallel Task"""
  def __init__(self, config, task_name, tokenizers, label_mapping):
    super().__init__(config, task_name, tokenizers, label_mapping)

  @property
  def example_class(self):
    return TwinsExample

  @property
  def minibatch_class(self):
    return TwinsMiniBatch

  def process_example(self, example: TwinsExample):
    example.process(tokenizers=self.tokenizers, label_mapping=self.label_mapping)

  def convert_examples_to_features(self, examples: List[TwinsExample]):
    examples: List[TwinsExample]
    features = []
    try:
      a_max_token_length = max([example.len_a for example in examples])
      b_max_token_length = max([example.len_b for example in examples])
    except:
      a_max_token_length = None
      b_max_token_length = None

    try:
      _a_max_token_length = max([example._len_a for example in examples])
      _b_max_token_length = max([example._len_b for example in examples])
    except:
      _a_max_token_length = None
      _b_max_token_length = None

    for idx, example in enumerate(examples):
      if a_max_token_length is not None:
        a_padding = [0] * (a_max_token_length - example.len_a)
        b_padding = [0] * (b_max_token_length - example.len_b)
        a_input_ids = example.a_input_ids + [example.padding_id] * (a_max_token_length - example.len_a)
        b_input_ids = example.b_input_ids + [example.padding_id] * (b_max_token_length - example.len_b)
        a_segment_ids = example.a_segment_ids + a_padding
        b_segment_ids = example.b_segment_ids + b_padding
        a_input_mask = example.a_input_mask + a_padding
        b_input_mask = example.b_input_mask + b_padding

        if example.label_ids is not None:
          # We assume the label length is same as sequence length
          label_ids = example.label_ids
        else:
          label_ids = None
      else:
        (a_input_ids, b_input_ids, a_segment_ids, b_segment_ids,
         a_input_mask, b_input_mask, label_ids) = None, None, None, None, None, None, None

      if _a_max_token_length is not None:
        _a_input_token_ids = example._a_input_token_ids + [0] * (_a_max_token_length - example._len_a)
        _a_token_length = example._len_a
        _a_input_token_mask = example._a_input_token_mask + [False] * (_a_max_token_length - example._len_a)
        _b_input_token_ids = example._b_input_token_ids + [0] * (_b_max_token_length - example._len_b)
        _b_token_length = example._len_b
        _b_input_token_mask = example._b_input_token_mask + [False] * (_b_max_token_length - example._len_b)
        if example._label_ids is not None:
          _label_ids = example._label_ids
        else:
          _label_ids = None
      else:
        (_a_input_token_ids, _b_input_token_ids, _a_token_length,
         _b_token_length, _a_input_token_mask, _b_input_token_mask, _label_ids) = None, None, None, None, None, None, None

      features.append(
        TwinsFeature(
          a_input_ids=a_input_ids,
          b_input_ids=b_input_ids,
          a_input_mask=a_input_mask,
          b_input_mask=b_input_mask,
          a_segment_ids=a_segment_ids,
          b_segment_ids=b_segment_ids,
          label_ids=label_ids,
          _a_input_token_ids=_a_input_token_ids,
          _a_input_token_mask=_a_input_token_mask,
          _a_token_length=_a_token_length,
          _b_input_token_ids=_b_input_token_ids,
          _b_input_token_mask=_b_input_token_mask,
          _b_token_length=_b_token_length,
          _label_ids=_label_ids))
    return features

  def decode_to_labels(self, preds, mbs: TwinsMiniBatch):
    return preds


from relogic.logickit.tasks.tagging import Tagging
from relogic.logickit.tasks.classification import Classification
from relogic.logickit.tasks.span_extraction import SpanExtraction, SpanGCN
from relogic.logickit.tasks.span_extraction import ECPExtraction
from relogic.logickit.tasks.parsing import Parsing
from relogic.logickit.tasks.unsupervised import Unsupervised
from relogic.logickit.base.constants import (ECP_TASK, IR_TASK, NER_TASK, PARALLEL_MAPPING_TASK,
                                             PARALLEL_TEACHER_STUDENT_TASK, PAIRWISE_TASK, ENTITY_TYPE_CLASSIFICATION,
                                             DEP_PARSING_TASK, MIXSENT_TASK, LANGUAGE_IDENTIFICATION_IR, POS_TASK, DOCIR_TASK,
                                             LANGUAGE_IDENTIFICATION_SEQ, PIPE_SRL_TASK, JOINT_SRL_TASK, PREDICATE_DETECTION_TASK)
from relogic.common.prefix_map import PrefixMap

task_name_to_class_map = {
  POS_TASK: Tagging,
  NER_TASK: Tagging,
  PREDICATE_DETECTION_TASK: Tagging,
  PIPE_SRL_TASK: Tagging,
  JOINT_SRL_TASK: Tagging,
  IR_TASK: Classification,
  DOCIR_TASK: Classification,
  PAIRWISE_TASK: Classification,
  ENTITY_TYPE_CLASSIFICATION: Classification,
  ECP_TASK: ECPExtraction,
  PARALLEL_MAPPING_TASK: Unsupervised,
  PARALLEL_TEACHER_STUDENT_TASK: Unsupervised,
  MIXSENT_TASK: Unsupervised,
  DEP_PARSING_TASK:Parsing,
  LANGUAGE_IDENTIFICATION_IR: Unsupervised,
  LANGUAGE_IDENTIFICATION_SEQ: Unsupervised,
}

TASK_NAME_TO_CLASS_MAP = PrefixMap(task_name_to_class_map)

def get_task(config, name, tokenizer):
  return TASK_NAME_TO_CLASS_MAP[name](config, name, tokenizer)

# def get_task(config, name, tokenizer):
#   if name in ["ccg", "pos"]:
#     return Tagging(config, name, True, tokenizer)
#   elif name in ["chunk", "ner",
#                 "er", "nestedner", "srl",
#                 "srl_conll05", "srl_conll09",
#                 "srl_conll12", "predicate_sense", "joint_srl", NER_TASK]:
#     return Tagging(config, name, False, tokenizer)
#   elif name in ["matching", "rel_extraction", "pair_matching", IR_TASK, PAIRWISE_TASK, ENTITY_TYPE_CLASSIFICATION]:
#     return Classification(config, name, tokenizer)
#   elif name in ["squad11", "squad20"]:
#     return SpanExtraction(config, name, tokenizer)
#   elif name in ["span_gcn"]:
#     return SpanGCN(config, name, tokenizer)
#   elif name in [ECP_TASK]:
#     return ECPExtraction(config, name, tokenizer)
#   elif name in [PARALLEL_MAPPING_TASK, PARALLEL_TEACHER_STUDENT_TASK, MIXSENT_TASK]:
#     return Unsupervised(config, name, tokenizer)
#   elif name in [DEP_PARSING_TASK]:
#     return Parsing(config, name, tokenizer)
#   else:
#     raise ValueError("Unknow task", name)


from relogic.logickit.dataset.labeled_data_loader import LabeledDataLoader
from relogic.logickit.dataflow import DataFlow

from relogic.logickit.base.constants import (
  NER_TASK
)
from relogic.logickit.modules.sequence_labeling_module import SequenceLabelingModule
from relogic.logickit.modules.matching_module import MatchingModule

# from relogic.logickit.scorer.ranking_scorer import RetrievalScorer
# import json
#
#
# task_name_to_scorer = {
#   IR_TASK: RetrievalScorer,# ,(self.loader.label_mapping, qrels_file_path=self.config.qrels_file_path, dump_to_file=dump_to_file)
# }
#
# TASK_NAME_TO_SCORER = PrefixMap(task_name_to_scorer)
#
# task_name_to_loss_func = {}
#
# TASK_NAME_TO_LOSS_FUNC = PrefixMap(task_name_to_loss_func)
#
# class Task(object):
#   """Unified Task Class
#   """
#   def __init__(self, name, config, task_config, tokenizers=None, data_loader=None):
#     self.config = config
#     self.name = name
#     self.loader = data_loader if data_loader is not None else LabeledDataLoader(config, name, tokenizers)
#     if config.mode == 'train' or config.mode == 'finetune':
#       self.train_set = self.loader.get_dataset("train")
#     else:
#       self.train_set = None
#     if config.mode != "deployment":
#       self.val_set = self.loader.get_dataset("dev" if (
#             config.mode == 'train' or config.mode == 'valid' or config.mode == "finetune") else "test")
#     else:
#       self.val_set = None
#     try:
#       self.dataset: DataFlow = self.loader.get_dataflow()
#     except:
#       self.dataset = None
#
#
#     self.task_config = json.load(open(config.module_configs[name]))
#     self.module = module if module is not None else TASK_NAME_TO_MODULE[name](config=config, **self.task_config)
#     # self.loss_func = TASK_NAME_TO_LOSS_FUNC[name](config=config, **self.task_config)
#
#   @property
#   def get_module(self):
#     return self.module
#
#   @classmethod
#   def from_config(cls, task_name, config=None, tokenizers=None, data_loader=None, task_config=None):
#     return cls(name=task_name, config=config, task_config=task_config, tokenizers=tokenizers, data_loader=data_loader)
#
#
#   def get_scorer(self):
#     return TASK_NAME_TO_SCORER[self.name](self.task_config)





import torch.nn as nn
import torch
from relogic.logickit.utils import utils


class EndpointSpanExtractor(nn.Module):
  def __init__(self,
               input_dim: int,
               combination: str = "x,y",
               num_width_embeddings: int = None,
               span_width_embedding_dim: int = None,
               bucket_widths: bool = False) -> None:
    super(EndpointSpanExtractor, self).__init__()
    self._input_dim = input_dim
    self._combination = combination
    self._num_width_embeddings = num_width_embeddings
    self._bucket_widths = bucket_widths

    if num_width_embeddings is not None and span_width_embedding_dim is not None:
      self._span_width_embedding = nn.Embedding(num_width_embeddings, span_width_embedding_dim)
    elif not all([num_width_embeddings is None, span_width_embedding_dim is None]):
      raise ValueError("To use a span width embedding representation, you must"
                               "specify both num_width_embeddings and span_width_embedding_dim.")
    else:
      self._span_width_embedding = None


  def forward(self,
              sequence_tensor: torch.FloatTensor,
              span_indices: torch.LongTensor,
              sequence_mask: torch.LongTensor = None,
              span_indices_mask: torch.LongTensor = None):
    span_starts, span_ends = [index.squeeze(-1) for index in span_indices.split(1, dim=-1)]

    if span_indices_mask is not None:
      span_starts = span_starts * span_indices_mask.long()
      span_ends = span_ends * span_indices_mask.long()

    # The span is exclusive on the right, so the span_ends need to -1

    start_embeddings = utils.batched_index_select(sequence_tensor, span_starts)
    inclusive_span_ends = torch.relu((span_ends - 1).float()).long()
    end_embeddings = utils.batched_index_select(sequence_tensor ,inclusive_span_ends)

    combined_tensors = torch.cat([start_embeddings, end_embeddings], dim=-1)

    if self._span_width_embedding is not None:
      # Embed the span widths and concatenate to the rest of the representations.
      if self._bucket_widths:
        span_widths = utils.bucket_values(span_ends - span_starts,
                                          num_total_buckets=self._num_width_embeddings)
      else:
        span_widths = span_ends - span_starts

      span_width_embeddings = self._span_width_embedding(span_widths)
      combined_tensors = torch.cat([combined_tensors, span_width_embeddings], dim=-1)

    if span_indices_mask is not None:
      return combined_tensors * span_indices_mask.unsqueeze(-1).float()

    return combined_tensors







import torch.nn as nn
import torch

class BilinearMatrixAttention(nn.Module):
  """
  Adopted from AllenNLP. For now there is no activation function
  """
  def __init__(
        self,
        matrix_1_dim: int,
        matrix_2_dim: int,
        use_input_biases: bool = False,
        label_dim: int = 1) -> None:
    super().__init__()
    if use_input_biases:
      matrix_1_dim += 1
      matrix_2_dim += 1

    if label_dim == 1:
      self.weight_matrix = nn.Parameter(torch.Tensor(matrix_1_dim, matrix_2_dim))
    else:
      self.weight_matrix = nn.Parameter(torch.Tensor(label_dim, matrix_1_dim, matrix_2_dim))
    self.bias = nn.Parameter(torch.Tensor(1))
    self.use_input_biases = use_input_biases

    self.reset_parameters()

  def reset_parameters(self):
    torch.nn.init.xavier_uniform_(self.weight_matrix)
    self.bias.data.fill_(0)

  def forward(self, matrix_1: torch.Tensor, matrix_2: torch.Tensor) -> torch.Tensor:
    if self.use_input_biases:
      bias1 = matrix_1.new_ones(matrix_1.size()[:-1] + (1,))
      bias2 = matrix_2.new_ones(matrix_2.size()[:-1] + (1,))

      matrix_1 = torch.cat([matrix_1, bias1], -1)
      matrix_2 = torch.cat([matrix_2, bias2], -1)

    weight = self.weight_matrix
    if weight.dim() == 2:
      weight = weight.unsqueeze(0)
    intermediate = torch.matmul(matrix_1.unsqueeze(1), weight)
    final = torch.matmul(intermediate, matrix_2.unsqueeze(1).transpose(2, 3))
    return final.squeeze(1) + self.bias






import argparse
from sklearn.decomposition import PCA
import numpy as np

pca = PCA(2)
RS = 123
from sklearn.manifold import TSNE


parser = argparse.ArgumentParser()
parser.add_argument("--vector_file_paths")
parser.add_argument("--output_file_path")
args = parser.parse_args()

files = args.vector_file_paths.split(",")
Xs = []
ys = []
boundary = []
l = 0
for idx, file in enumerate(files):
  fin = open(file, 'rb')
  data = np.load(fin)
  Xs.append(data['X'])
  if ['y'] in data.files:
    ys.append(data['y'])
  else:
    ys.append(np.ones(data['X'].shape[0], int) * idx)
  boundary.append((l, l+len(data['X'])))
  l += len(data['X'])
X = np.concatenate(Xs, 0)
y = np.concatenate(ys, 0)

X_PCA = pca.fit_transform(X)
X_proj = fashion_tsne = TSNE(random_state=RS).fit_transform(X_PCA)

import matplotlib.pyplot as plt

fig, axes = plt.subplots(1, len(files) + 1)
for i in range(len(files)):
  b = boundary[i]
  axes[i].scatter(X_proj[b[0]:b[1],0], X_proj[b[0]:b[1], 1], c=y[b[0]:b[1]], s=0.5)

axes[-1].scatter(X_proj[:,0], X_proj[:, 1], c=y, s=0.5)
plt.savefig(args.output_file_path)

print(pca.components_)
print(pca.components_.shape)


from dataclasses import dataclass
from typing import Dict, List

from relogic.structures.structure import Structure


@dataclass
class LinkageCandidate(Structure):
  """
  """
  text: str
  score: float
  prior: Dict
  alias_of: List
  
  @classmethod
  def from_hit(cls, hit):
    prior_table = {}
    if hit.prior:
      priors = hit.prior.strip().split("\t")
      for uri, count in zip(priors[::2], priors[1::2]):
        uri = uri.replace("WikiProject_", "")
        # TODO: We will unified the cleaning process and add it to utils
        prior_table[uri] = prior_table.get(uri, 0) + int(count)
    alias_of = [] if not hit.alias else hit.alias.split("\t")
    return cls(text=hit.content, score=hit.score, prior=prior_table, alias_of=alias_of)



import argparse
import math
def main(input_file, dump_file):
  with open(input_file) as fin:
    sentences_scores = []
    scores = []
    for line in open(input_file):
      if len(line.strip()) == 0:
        sentences_scores.append(scores)
        scores = []
        continue
      token, _, _, confidence = line.strip().split()
      scores.append(float(confidence))
  with open(dump_file, 'w') as fout:
    for idx, scores in enumerate(sentences_scores):
      sum_score = sum([math.log(score) for score in scores]) / len(scores)
      fout.write("{}\t{}\n".format(idx, sum_score))

if __name__ == "__main__":
  parser = argparse.ArgumentParser()
  parser.add_argument("--input_file")
  parser.add_argument("--dump_file")
  args = parser.parse_args()
  main(args.input_file, args.dump_file)


from relogic.logickit.tokenizer.fasttext_tokenization import FasttextTokenizer

tokenizer = FasttextTokenizer.from_pretrained("wiki-news-300d-1M")
text = "take by surprise, taking by surprise (or some other manner)"
print(tokenizer.tokenize(text))


from relogic.logickit.utils.utils import get_span_labels
import argparse
import json

def merge(gold_examples, predicted_examples):
  assert len(gold_examples) == len(predicted_examples)
  for gold_example, predicted_example in zip(gold_examples, predicted_examples):
    predicted_spans, labels = get_span_labels(predicted_example["labels"])
    sorted_spans = sorted(predicted_spans, key=lambda span: span[0], reverse=False)
    predicted_spans = [(span[0], span[1]) for span in sorted_spans]
    predicted_spans_label = [span[2] for span in sorted_spans]
    gold_example["span_candidates"] = predicted_spans
    gold_example["span_candidates_label"] = predicted_spans_label


if __name__ == "__main__":
  parser = argparse.ArgumentParser()
  parser.add_argument("--gold_data_path", type=str)
  parser.add_argument("--pred_data_path", type=str)
  parser.add_argument("--output_data_path", type=str)
  args = parser.parse_args()
  with open(args.gold_data_path) as fin:
    gold_examples = []
    for line in fin:
      gold_examples.append(json.loads(line))
  with open(args.pred_data_path) as fin:
    predicted_examples = []
    for line in fin:
      predicted_examples.append(json.loads(line))
  merge(gold_examples, predicted_examples)
  with open(args.output_data_path, "w") as fout:
    for example in gold_examples:
      fout.write(json.dumps(example) + "\n")
  




import tensorflow as tf
import csv

class DataProcessor(object):
  """Base class for data converters for sequence classification data sets."""

  def get_train_examples(self, data_dir):
    """Gets a collection of `InputExample`s for the train set."""
    raise NotImplementedError()

  def get_dev_examples(self, data_dir):
    """Gets a collection of `InputExample`s for the dev set."""
    raise NotImplementedError()

  def get_test_examples(self, data_dir):
    """Gets a collection of `InputExample`s for prediction."""
    raise NotImplementedError()

  def get_labels(self):
    """Gets the list of labels for this data set."""
    raise NotImplementedError()

  @classmethod
  def _read_tsv(cls, input_file, quotechar=None):
    """Reads a tab separated value file."""
    with tf.gfile.Open(input_file, "r") as f:
      reader = csv.reader(f, delimiter="\t", quotechar=quotechar)
      lines = []
      for line in reader:
        lines.append(line)
      return lines


import argparse
import json

def main(json_file, output_file):
  fout = open(output_file, 'w')
  with open(json_file) as fin:
    for line in fin:
      example = json.loads(line)
      relation = example["relation"]
      sub = example["sub"]
      obj = example["obj"]
      tokens = example["tokens"]
      fout.write("#Rel#\t{}\n".format(relation))
      fout.write("#Subj#\t{}\t{}\t{}\t{}\n".format(
        sub["start"],
        sub["end"],
        sub["text"],
        sub["entityType"]))
      fout.write("#Obj#\t{}\t{}\t{}\t{}\n".format(
        obj["start"],
        obj["end"],
        obj["text"],
        obj["entityType"]))
      for token in tokens:
        fout.write("{}\n".format(token))
      fout.write("\n")

if __name__ == "__main__":
  parser = argparse.ArgumentParser()
  parser.add_argument("--json_file", type=str)
  parser.add_argument("--output_file", type=str)
  args = parser.parse_args()
  main(args.json_file, args.output_file)


import torch
import relogic.utils.crash_on_ipy

def clean(text):
  return text.strip()

class RobertaXLMTokenizer:
  def __init__(self, model):
    self.roberta = torch.hub.load('pytorch/fairseq', model)

  @classmethod
  def from_pretrained(cls, pretrained_model_name_or_path):
    return cls(model=pretrained_model_name_or_path)

  # def special_match(self, other_tok, bpe_tok):
  #   if other_tok == '΄' and bpe_tok == "́":
  #     return True
  #   return False

  def alignment(self, bpe_tokens, expected_tokenization):
    bpe_tokens = [clean(text) for text in bpe_tokens]
    alignment = []
    bpe_toks_list = list(filter(lambda item: item[1] != '', enumerate(bpe_tokens, start=1)))
    bpe_toks = filter(lambda item: item[1] != '', enumerate(bpe_tokens, start=1))
    j, bpe_tok = next(bpe_toks)
    for other_tok in expected_tokenization:
      other_tok = other_tok.replace('΄', "́")
      bpe_indices = []
      while True:
        if other_tok.startswith(bpe_tok):
          bpe_indices.append(j)
          other_tok = other_tok[len(bpe_tok):]
          try:
            j, bpe_tok = next(bpe_toks)
          except StopIteration:
            j, bpe_tok = None, None
        elif bpe_tok.startswith(other_tok):
          # other_tok spans multiple BPE tokens
          bpe_indices.append(j)
          bpe_tok = bpe_tok[len(other_tok):]
          other_tok = ''
        else:
          print(bpe_tokens)
          print(bpe_toks_list)
          print(expected_tokenization)
          raise Exception('Cannot align "{}" and "{}"'.format(other_tok, bpe_tok))
        if other_tok == '':
          break
      assert len(bpe_indices) > 0
      alignment.append(bpe_indices)
    assert len(alignment) == len(expected_tokenization)
    return alignment

  def tokenize_and_add_placeholder_and_convert_to_ids(self, sent, expected_tokenization):
    roberta_ids = self.roberta.encode(sent)
    bpe_tokens = [self.roberta.bpe.decode(self.roberta.task.source_dictionary.string([x])) for x in roberta_ids]

    if expected_tokenization is not None:
      alignmet = self.alignment(bpe_tokens, expected_tokenization)
      is_head = [2]
      for word in alignmet:
        is_head.append(1)
        for _ in word[1:]:
          is_head.append(0)
      is_head.append(2)

    # Fix for empty
    fixed_bpe_tokens = []
    fixed_roberta_ids = []

    for idx, (token, roberta_id) in enumerate(zip(bpe_tokens, roberta_ids.tolist())):

      if token.strip() != "" or idx == 0 or idx == len(bpe_tokens)-1:
        fixed_bpe_tokens.append(token)
        fixed_roberta_ids.append(roberta_id)

    if expected_tokenization is not None:
      assert(len(fixed_bpe_tokens) == len(is_head))
    else:
      is_head = [0] * len(fixed_bpe_tokens)

    return fixed_bpe_tokens, is_head, fixed_roberta_ids

if __name__ == "__main__":
  sent = " ".join(['Μεγάλης', 'κλίμακας', 'αναστηλωτικές', 'εργασίες', 'έγιναν', 'από', 'τη', 'Γαλλική',
                   'Αρχαιολογική', 'Σχολή', ',', 'κυρίως', 'στις', 'συνοικίες', 'ιδιωτικών', 'κατοικιών',
                   ',', 'ενώ', 'μικρότερης', 'κλίμακας', 'πραγματοποιούνται', 'τα', 'τελευταία', 'χρόνια',
                   'από', 'την', 'ΚΑ΄', 'Εφορεία', 'Προϊστορικών', 'και', 'Κλασικών', 'Αρχαιοτήτων', '.'])
  tokenizer = RobertaXLMTokenizer("xlmr.large.v0")
  raise NotImplementedError("Test")


from typing import List

import torch

from relogic.logickit.dataflow import DataFlow, Example, Feature, MiniBatch
from relogic.logickit.tokenizer.tokenization import BertTokenizer
from relogic.logickit.utils import create_tensor, truncate_seq_pair


class PairwiseExample(Example):
  """
  text (str): query text
  text_p (str): positive example
  text_n (str): negative example
  """
  def __init__(self, guid: str, text: str, p_guid: str=None, text_p:str=None, n_guid: str=None, text_n:str=None):
    super(PairwiseExample, self).__init__()
    self.guid = guid
    self.text = text
    self.p_guid = p_guid
    self.text_p = text_p
    self.n_guid = n_guid
    self.text_n = text_n

  def process(self, tokenizers, *inputs, **kwargs):
    for tokenizer in tokenizers.values():
      if isinstance(tokenizer, BertTokenizer):
        max_seq_length = kwargs.pop("max_seq_length", 512)

        self.text_tokens, self.text_is_head = tokenizer.tokenize(self.text)
        self.text_tokens = self.text_tokens[:max_seq_length - 2]
        self.text_is_head = self.text_is_head[:max_seq_length - 2]
        self.tokens = ["[CLS]"] + self.text_tokens + ["[SEP]"]
        self.input_ids = tokenizer.convert_tokens_to_ids(self.tokens)
        self.input_mask = [1] * len(self.input_ids)
        self.segment_ids = [0] * len(self.input_ids)

        if self.text_p:
          self.text_p_tokens, self.text_p_is_head = tokenizer.tokenize(self.text_p)
          self.text_p_tokens = self.text_p_tokens[:max_seq_length - 2]
          self.text_p_is_head = self.text_p_is_head[:max_seq_length - 2]
          self.p_tokens = ["[CLS]"] + self.text_p_tokens + ["[SEP]"]
          self.p_input_ids = tokenizer.convert_tokens_to_ids(self.p_tokens)
          self.p_input_mask = [1] * len(self.p_input_ids)
          self.p_segment_ids = [0] * len(self.p_input_ids)

        if self.text_n:
          self.text_n_tokens, self.text_n_is_head = tokenizer.tokenize(self.text_n)
          self.text_n_tokens = self.text_n_tokens[:max_seq_length - 2]
          self.text_n_is_head = self.text_n_is_head[:max_seq_length - 2]
          self.n_tokens = ["[CLS]"] + self.text_n_tokens + ["[SEP]"]
          self.n_input_ids = tokenizer.convert_tokens_to_ids(self.n_tokens)
          self.n_input_mask = [1] * len(self.n_input_ids)
          self.n_segment_ids = [0] * len(self.n_input_ids)

  @classmethod
  def from_structure(cls, structure):
    return cls(guid="", text=structure.text)

  @classmethod
  def from_json(cls, example):
    return cls(guid=example.get("guid", None),
               text=example["text"],
               text_p=example.get("text_p", None),
               text_n=example.get("text_n", None),
               p_guid=example.get("p_guid", None),
               n_guid=example.get("n_guid", None))

  @property
  def len(self):
    return len(self.tokens)

  @property
  def len_p(self):
    return len(self.p_tokens)

  @property
  def len_n(self):
    return len(self.n_tokens)

class PairwiseFeature(Feature):
  """

  """
  def __init__(self, *inputs, **kwargs):
    super(PairwiseFeature, self).__init__()
    # BERT based feature
    self.input_ids = kwargs.pop("input_ids")
    self.input_mask = kwargs.pop("input_mask")
    self.segment_ids = kwargs.pop("segment_ids")
    self.p_input_ids = kwargs.pop("p_input_ids", None)
    self.p_input_mask = kwargs.pop("p_input_mask", None)
    self.p_segment_ids = kwargs.pop("p_segment_ids", None)
    self.n_input_ids = kwargs.pop("n_input_ids", None)
    self.n_input_mask = kwargs.pop("n_input_mask", None)
    self.n_segment_ids = kwargs.pop("n_segment_ids", None)

class PairwiseMiniBatch(MiniBatch):
  def __init__(self, *inputs, **kwargs):
    super(PairwiseMiniBatch, self).__init__(*inputs, **kwargs)

  def generate_input(self, device, use_label):
    inputs = {}
    inputs["task_name"] = self.task_name
    inputs["input_ids"] = create_tensor(self.input_features, "input_ids",
                                        torch.long, device)
    inputs["input_mask"] = create_tensor(self.input_features, "input_mask",
                                        torch.long, device)
    inputs["segment_ids"] = create_tensor(self.input_features, "segment_ids",
                                          torch.long, device)
    inputs["p_input_ids"] = create_tensor(self.input_features, "p_input_ids",
                                          torch.long, device)
    inputs["p_input_mask"] = create_tensor(self.input_features, "p_input_mask",
                                           torch.long, device)
    inputs["p_segment_ids"] = create_tensor(self.input_features, "p_segment_ids",
                                            torch.long, device)
    inputs["n_input_ids"] = create_tensor(self.input_features, "n_input_ids",
                                          torch.long, device)
    inputs["n_input_mask"] = create_tensor(self.input_features, "n_input_mask",
                                           torch.long, device)
    inputs["n_segment_ids"] = create_tensor(self.input_features, "n_segment_ids",
                                            torch.long, device)
    inputs["is_inference"] = not use_label
    inputs["extra_args"] = {"target": -torch.ones(inputs["input_ids"].size(0)).to(device)}

    return inputs

class PairwiseDataFlow(DataFlow):
  def __init__(self, config, task_name, tokenizers, label_mapping=None):
    super(PairwiseDataFlow, self).__init__(config, task_name, tokenizers, label_mapping)

  @property
  def example_class(self):
    return PairwiseExample

  @property
  def minibatch_class(self):
    return PairwiseMiniBatch

  def process_example(self, example: PairwiseExample):
    example.process(tokenizers=self.tokenizers, max_seq_length=self.config.max_seq_length)

  def convert_examples_to_features(self, examples: List[PairwiseExample]):
    examples: List[PairwiseExample]
    features = []

    max_token_length = max([example.len for example in examples])
    try:
      max_token_length_p = max([example.len_p for example in examples])
      max_token_length_n = max([example.len_n for example in examples])
    except:
      max_token_length_p = 0
      max_token_length_n = 0

    for idx, example in enumerate(examples):
      padding = [0] * (max_token_length - example.len)
      input_ids = example.input_ids + padding
      input_mask = example.input_mask + padding
      segment_ids = example.segment_ids + padding

      if max_token_length_p > 0 and max_token_length_n > 0:
        p_padding = [0] * (max_token_length_p - example.len_p)
        p_input_ids = example.p_input_ids + p_padding
        p_input_mask = example.p_input_mask + p_padding
        p_segment_ids = example.p_segment_ids + p_padding

        n_padding = [0] * (max_token_length_n - example.len_n)
        n_input_ids = example.n_input_ids + n_padding
        n_input_mask = example.n_input_mask + n_padding
        n_segment_ids = example.n_segment_ids + n_padding

        features.append(
          PairwiseFeature(input_ids=input_ids,
                          input_mask=input_mask,
                          segment_ids=segment_ids,
                          p_input_ids=p_input_ids,
                          p_input_mask=p_input_mask,
                          p_segment_ids=p_segment_ids,
                          n_input_ids=n_input_ids,
                          n_input_mask=n_input_mask,
                          n_segment_ids=n_segment_ids))
      else:
        features.append(
          PairwiseFeature(input_ids=input_ids,
                          input_mask=input_mask,
                          segment_ids=segment_ids))
    return features








from relogic.tpukit.processors.data_processor import DataProcessor
import os
from relogic.tpukit.tokenizer import tokenization
from relogic.tpukit.input_fn import InputExample


class XnliProcessor(DataProcessor):
  """Processor for the XNLI data set."""

  def __init__(self):
    self.language = "zh"

  def get_train_examples(self, data_dir):
    """See base class."""
    lines = self._read_tsv(
        os.path.join(data_dir, "multinli",
                     "multinli.train.%s.tsv" % self.language))
    examples = []
    for (i, line) in enumerate(lines):
      if i == 0:
        continue
      guid = "train-%d" % (i)
      text_a = tokenization.convert_to_unicode(line[0])
      text_b = tokenization.convert_to_unicode(line[1])
      label = tokenization.convert_to_unicode(line[2])
      if label == tokenization.convert_to_unicode("contradictory"):
        label = tokenization.convert_to_unicode("contradiction")
      examples.append(
          InputExample(guid=guid, text_a=text_a, text_b=text_b, label=label))
    return examples

  def get_dev_examples(self, data_dir):
    """See base class."""
    lines = self._read_tsv(os.path.join(data_dir, "xnli.dev.tsv"))
    examples = []
    for (i, line) in enumerate(lines):
      if i == 0:
        continue
      guid = "dev-%d" % (i)
      language = tokenization.convert_to_unicode(line[0])
      if language != tokenization.convert_to_unicode(self.language):
        continue
      text_a = tokenization.convert_to_unicode(line[6])
      text_b = tokenization.convert_to_unicode(line[7])
      label = tokenization.convert_to_unicode(line[1])
      examples.append(
          InputExample(guid=guid, text_a=text_a, text_b=text_b, label=label))
    return examples

  def get_labels(self):
    """See base class."""
    return ["contradiction", "entailment", "neutral"]

class MnliProcessor(DataProcessor):
  """Processor for the MultiNLI data set (GLUE version)."""

  def get_train_examples(self, data_dir):
    """See base class."""
    return self._create_examples(
        self._read_tsv(os.path.join(data_dir, "train.tsv")), "train")

  def get_dev_examples(self, data_dir):
    """See base class."""
    return self._create_examples(
        self._read_tsv(os.path.join(data_dir, "dev_matched.tsv")),
        "dev_matched")

  def get_test_examples(self, data_dir):
    """See base class."""
    return self._create_examples(
        self._read_tsv(os.path.join(data_dir, "test_matched.tsv")), "test")

  def get_labels(self):
    """See base class."""
    return ["contradiction", "entailment", "neutral"]

  def _create_examples(self, lines, set_type):
    """Creates examples for the training and dev sets."""
    examples = []
    for (i, line) in enumerate(lines):
      if i == 0:
        continue
      guid = "%s-%s" % (set_type, tokenization.convert_to_unicode(line[0]))
      text_a = tokenization.convert_to_unicode(line[8])
      text_b = tokenization.convert_to_unicode(line[9])
      if set_type == "test":
        label = "contradiction"
      else:
        label = tokenization.convert_to_unicode(line[-1])
      examples.append(
          InputExample(guid=guid, text_a=text_a, text_b=text_b, label=label))
    return examples


class MrpcProcessor(DataProcessor):
  """Processor for the MRPC data set (GLUE version)."""

  def get_train_examples(self, data_dir):
    """See base class."""
    return self._create_examples(
        self._read_tsv(os.path.join(data_dir, "train.tsv")), "train")

  def get_dev_examples(self, data_dir):
    """See base class."""
    return self._create_examples(
        self._read_tsv(os.path.join(data_dir, "dev.tsv")), "dev")

  def get_test_examples(self, data_dir):
    """See base class."""
    return self._create_examples(
        self._read_tsv(os.path.join(data_dir, "test.tsv")), "test")

  def get_labels(self):
    """See base class."""
    return ["0", "1"]

  def _create_examples(self, lines, set_type):
    """Creates examples for the training and dev sets."""
    examples = []
    for (i, line) in enumerate(lines):
      if i == 0:
        continue
      guid = "%s-%s" % (set_type, i)
      text_a = tokenization.convert_to_unicode(line[3])
      text_b = tokenization.convert_to_unicode(line[4])
      if set_type == "test":
        label = "0"
      else:
        label = tokenization.convert_to_unicode(line[0])
      examples.append(
          InputExample(guid=guid, text_a=text_a, text_b=text_b, label=label))
    return examples


class ColaProcessor(DataProcessor):
  """Processor for the CoLA data set (GLUE version)."""

  def get_train_examples(self, data_dir):
    """See base class."""
    return self._create_examples(
        self._read_tsv(os.path.join(data_dir, "train.tsv")), "train")

  def get_dev_examples(self, data_dir):
    """See base class."""
    return self._create_examples(
        self._read_tsv(os.path.join(data_dir, "dev.tsv")), "dev")

  def get_test_examples(self, data_dir):
    """See base class."""
    return self._create_examples(
        self._read_tsv(os.path.join(data_dir, "test.tsv")), "test")

  def get_labels(self):
    """See base class."""
    return ["0", "1"]

  def _create_examples(self, lines, set_type):
    """Creates examples for the training and dev sets."""
    examples = []
    for (i, line) in enumerate(lines):
      # Only the test set has a header
      if set_type == "test" and i == 0:
        continue
      guid = "%s-%s" % (set_type, i)
      if set_type == "test":
        text_a = tokenization.convert_to_unicode(line[1])
        label = "0"
      else:
        text_a = tokenization.convert_to_unicode(line[3])
        label = tokenization.convert_to_unicode(line[1])
      examples.append(
          InputExample(guid=guid, text_a=text_a, text_b=None, label=label))
    return examples





import json
from types import SimpleNamespace

import relogic.utils.crash_on_ipy
from relogic.logickit.base.constants import SEQUENCE_LABELING_TASK
from relogic.logickit.dataflow import TASK_TO_DATAFLOW_CLASS_MAP, SequenceDataFlow
from transformers.tokenization_bert import BertTokenizer

config = SimpleNamespace(
  **{
    "buckets": [(0, 15), (15, 40), (40, 450)],
    "max_seq_length": 450,
    "label_mapping_path": "data/preprocessed_data/er_BIOES_label_mapping.json",
    "tasks" : {
      SEQUENCE_LABELING_TASK: {
        "selected_non_final_layers": None}}})

tokenizers = {
  "BPE": BertTokenizer.from_pretrained("bert-base-multilingual-cased", do_lower_case=False,
                                       lang="zh", pretokenized=True),
}

dataflow: SequenceDataFlow = TASK_TO_DATAFLOW_CLASS_MAP[SEQUENCE_LABELING_TASK](
  task_name=SEQUENCE_LABELING_TASK,
  config=config,
  tokenizers=tokenizers,
  label_mapping=json.load(open(config.label_mapping_path)))

examples = [{
  "tokens": ["I", "visited", "China", "yesterday", "."],
  "labels": ["O", "O", "S-LOC", "O", "O"]
},{
  "tokens": ["Barack", "Obama", "went", "to", "Paris", "."],
  "labels": ["B-PERSON", "E-PERSON", "O", "O", "S-LOC", "O"]
}]
# examples = [ {
#  "tokens": [
#   "哲人",
#   "已",
#   "远",
#   "，",
#   "典范",
#   "长",
#   "存",
#   "——",
#   "悼",
#   "许常惠",
#   "、",
#   "张光直",
#   "与",
#   "戴国辉"
#  ],
#  "labels": [
#   "O",
#   "O",
#   "O",
#   "O",
#   "O",
#   "O",
#   "O",
#   "O",
#   "O",
#   "B-PERSON",
#   "O",
#   "B-PERSON",
#   "O",
#   "B-PERSON"
#  ]
# }, {
#  "tokens": [
#   "（",
#   "李光真",
#   "）"
#  ],
#  "labels": [
#   "O",
#   "B-PERSON",
#   "O"
#  ]
# }]

dataflow.update_with_jsons(examples)

for mb in dataflow.get_minibatches(minibatch_size=2):
  print(mb)

raise NotImplementedError("You can start to play with data")


from typing import List

import torch

from relogic.logickit.dataflow import DataFlow, Example, Feature, MiniBatch
from relogic.logickit.tokenizer.tokenization import BertTokenizer
from relogic.logickit.utils import create_tensor, truncate_seq_pair


class PointwiseExample(Example):
  """Pointwise Example contains the attributes and functionality of
  a pointwise example. This class is mainly used for sentence pair
  classification task, such as information retrieval, sentence similarity,
  with cross-entropy loss function.

  In this example, we call the first sentence as text_a, and the second sentence
  as text_b.

  Args:
    guid (str): global unique identifier for this pair. This information is usually used
      in the ranking problem.
    text (str):
  """
  def __init__(self, guid: str, text_a: str, text_b:str, label=None):
    super(PointwiseExample, self).__init__()
    self.guid = guid
    self.text_a = text_a
    self.text_b = text_b
    self.label = label

  def process(self, tokenizers, *inputs, regression=False, **kwargs):
    """Process the sentence pair.

    :param tokenizers:
    :param inputs:
    :param kwargs:
    :return:
    """
    for tokenizer in tokenizers.values():
      if isinstance(tokenizer, BertTokenizer):
        self.text_a_tokens, self.text_a_is_head = tokenizer.tokenize(self.text_a)
        self.text_b_tokens, self.text_b_is_head = tokenizer.tokenize(self.text_b)

        max_seq_length = kwargs.pop("max_seq_length", 512)

        truncate_seq_pair(self.text_a_tokens, self.text_b_tokens, max_seq_length - 3)
        truncate_seq_pair(self.text_a_is_head, self.text_b_is_head, max_seq_length - 3)

        self.tokens = ["[CLS]"] + self.text_a_tokens + ["[SEP]"] + self.text_b_tokens + ["[SEP]"]
        self.segment_ids = [0] * (len(self.text_a_tokens) + 2) + [1] * (len(self.text_b_tokens) + 1)
        self.is_head = [2] + self.text_a_is_head + [2] + self.text_b_is_head + [2]

        self.input_ids = tokenizer.convert_tokens_to_ids(self.tokens)
        self.input_mask = [1] * len(self.input_ids)

        self.text_a_indices = []
        self.text_b_indices = []
        for idx, ind in enumerate(self.text_a_is_head):
          if ind == 1:
            self.text_a_indices.append(idx + 1)
        for idx, ind in enumerate(self.text_b_is_head):
          if ind == 1:
            self.text_b_indices.append(idx + len(self.text_a_is_head) + 2)

        if self.label is not None:
          if regression:
            # Regression Problem with sigmoid loss function
            self.label_ids = float(self.label)
          else:
            label_mapping = kwargs.get("label_mapping")
            self.label_ids = label_mapping[self.label]


  @classmethod
  def from_structure(cls, structure):
    return cls(guid="", text_a=structure.text_a, text_b=structure.text_b)

  @classmethod
  def from_json(cls, example):
    """
    """
    if isinstance(example["text_a"], str) and isinstance(example["text_b"], str):
      return cls(guid="{}|{}".format(example.get("text_a_id", 0), example.get("text_b_id", 0)),
               text_a=example["text_a"],
               text_b=example["text_b"],
               label=example.get("label", None))
    else:
      return cls(guid="{}|{}".format(example.get("text_a_id", 0), example.get("text_b_id", 0)),
                 text_a=" ".join(example["text_a"]),
                 text_b=" ".join(example["text_b"]),
                 label=example.get("label", None))

  @property
  def len(self):
      return len(self.input_ids)


class PointwiseFeature(Feature):
  """

  """
  def __init__(self, *inputs, **kwargs):
    super(PointwiseFeature, self).__init__()
    # BERT based feature
    self.input_ids = kwargs.pop("input_ids")
    self.input_mask = kwargs.pop("input_mask")
    self.is_head = kwargs.pop("is_head")
    self.segment_ids = kwargs.pop("segment_ids")
    self.text_a_indices = kwargs.pop("text_a_indices")
    self.text_b_indices = kwargs.pop("text_b_indices")
    self.label_ids = kwargs.pop("label_ids")

class PointwiseMiniBatch(MiniBatch):
  """

  """
  def __init__(self, *inputs, **kwargs):
    super(PointwiseMiniBatch, self).__init__(*inputs, **kwargs)

  def generate_input(self, device, use_label):
    """Generate tensors based on PointwiseFeatures
    """
    # BERT based feature
    inputs = {}
    inputs["task_name"] = self.task_name
    inputs["input_ids"] = create_tensor(self.input_features, "input_ids",
                                        torch.long, device)
    inputs["input_mask"] = create_tensor(self.input_features, "input_mask",
                                         torch.long, device)
    inputs["segment_ids"] = create_tensor(self.input_features, "segment_ids",
                                          torch.long, device)
    inputs["input_head"] = create_tensor(self.input_features, "is_head",
                                          torch.long, device)
    inputs["text_a_indices"] = create_tensor(self.input_features, "text_a_indices",
                                             torch.long, device)
    inputs["text_b_indices"] = create_tensor(self.input_features, "text_b_indices",
                                             torch.long, device)
    if use_label:
      if self.config.regression:
        inputs["label_ids"] = create_tensor(self.input_features, "label_ids",
                                            torch.float, device)
      else:
        inputs["label_ids"] = create_tensor(self.input_features, "label_ids",
                                            torch.long, device)
    else:
      inputs["label_ids"] = None

    inputs["extra_args"] = {}
    if self.config.tasks[self.task_name]["selected_non_final_layers"] is not None:
      inputs["extra_args"]["selected_non_final_layers"] = self.config.tasks[self.task_name]["selected_non_final_layers"]

    return inputs

class PointwiseDataFlow(DataFlow):
  """DataFlow implementation based on Pointwise task"""
  def __init__(self, config, task_name, tokenizers, label_mapping):
    super(PointwiseDataFlow, self).__init__(config, task_name, tokenizers, label_mapping)

  @property
  def example_class(self):
    return PointwiseExample

  @property
  def minibatch_class(self):
    return PointwiseMiniBatch

  def process_example(self, example: PointwiseExample):
    """Process Pointwise example"""
    example.process(tokenizers=self.tokenizers,
                    label_mapping=self.label_mapping,
                    regression=self.config.regression)

  def convert_examples_to_features(self, examples: List[PointwiseExample]):
    examples: List[PointwiseExample]
    features = []

    # BERT based variables
    max_token_length = max([example.len for example in examples])
    max_text_a_indices_length = max([len(example.text_a_indices) for example in examples])
    max_text_b_indices_length = max([len(example.text_b_indices) for example in examples])
    for idx, example in enumerate(examples):
      # BERT based feature process
      padding = [0] * (max_token_length - example.len)
      input_ids = example.input_ids + padding
      input_mask = example.input_mask + padding
      segment_ids = example.segment_ids + padding
      is_head = example.is_head + padding
      a_indices_padding = [0] * (max_text_a_indices_length - len(example.text_a_indices))
      b_indices_padding = [0] * (max_text_b_indices_length - len(example.text_b_indices))
      text_a_indices = example.text_a_indices + a_indices_padding
      text_b_indices = example.text_b_indices + b_indices_padding
      if hasattr(example, "label_ids"):
        label_ids = example.label_ids
      else:
        label_ids = None

      features.append(
        PointwiseFeature(input_ids=input_ids,
                         input_mask=input_mask,
                         segment_ids=segment_ids,
                         is_head=is_head,
                         text_a_indices=text_a_indices,
                         text_b_indices=text_b_indices,
                         label_ids=label_ids))
    return features

  def decode_to_labels(self, preds, mbs: PointwiseMiniBatch):
    return preds


from dataclasses import dataclass, field
from typing import List, Dict, Union

from relogic.structures.structure import Structure
from relogic.structures.token import Token
from relogic.structures.linkage_candidate import LinkageCandidate

@dataclass
class Span(Structure):
  text: str = None
  _text: str = field(init=False, repr=False)
  tokens: List[Token] = field(default_factory=list)
  start_index: int = None
  end_index: int = None

  
  # Linking
  linkage_candidates: List[LinkageCandidate] = field(default_factory=list)
  aggregated_prior: Dict = field(default_factory=dict)
  first_layer_prior: Dict = field(default_factory=dict)
  ranked_uris: List = field(default_factory=list)
  wikipedia_uri: str = None

  def add_linkage_candidate(self, linkage: LinkageCandidate):
    """
    """
    self.linkage_candidates.append(linkage)
    for uri, count in linkage.prior.items():
      self.aggregated_prior[uri] = self.aggregated_prior.get(uri, 0) + count
  
  def add_token(self, token: Union[str, Token]):
    """
    """
    if isinstance(token, str):
      self.tokens.append(Token(text=token))
    else:
      self.tokens.append(token)
    if self._text is not None:
      # Because the sentence has been mutated, the text string
      # needs to be regenerated.
      self._text = None

  def set_start_index(self, start_index: int):
    self.start_index = start_index

  def set_end_index(self, end_index: int):
    self.end_index = end_index

  def add_wikipedia_uri(self, uri: str):
    self.wikipedia_uri = uri

  @property
  def text(self):
    if self._text is None:
      self._text = " ".join([token.text for token in self.tokens])
    return self._text
  
  @text.setter
  def text(self, text):
    if isinstance(text, property):
      # This is a ugly fix for non-arguments initialization
      # s = Span()
      # TODO: find better solution for this implementation
      self._text = None
    else:
      self._text = text


import torch
import torch.nn as nn

class AggMatchingModule(nn.Module):
  def __init__(self, config, task_name, n_classes):
    super().__init__()
    self.config = config
    self.task_name = task_name
    if self.config.regression:
      self.n_classes = 1
    else:
      self.n_classes = n_classes
    self.to_logits = nn.Linear(config.hidden_size, self.n_classes)

  # def forward(self, input, input_mask=None, segment_ids=None, extra_args=None, **kwargs):
  def forward(self, *inputs, **kwargs):
    features = kwargs.pop("features")
    features = features[:, 0]
    mean_feature = torch.mean(features, dim=0, keepdim=True)
    logits = self.to_logits(mean_feature)
    return logits




import json
from typing import List, Tuple, Dict
import os

import torch

from relogic.logickit.dataflow import DataFlow, Example, Feature, MiniBatch
from relogic.logickit.tokenizer.tokenization import BertTokenizer
from relogic.logickit.utils import create_tensor, filter_head_prediction


class SingletonExample(Example):
  """SingletonExample contains the attributes and functionality of an Sentence Classification example.

  Args:
    text (str): A sentence string.
    labels (str or List[str]): A label string or a list of label string.
  """
  def __init__(self, text, labels=None):
    super(SingletonExample, self).__init__()
    self.text = text
    self.raw_tokens = text.split()
    self.labels = labels

  def process(self, tokenizers: Dict, *inputs, **kwargs):
    """Process the Singleton Example..
    """
    for tokenizer in tokenizers.values():
      if isinstance(tokenizer, BertTokenizer):
        # BERT process part
        self.text_tokens, self.text_is_head = tokenizer.tokenize(self.text)
        self.tokens = ["[CLS]"] + self.text_tokens + ["[SEP]"]
        self.segment_ids = [0] * (len(self.tokens))
        self.is_head = [2] + self.text_is_head + [2]
        self.head_index = [idx for idx, value in enumerate(self.is_head) if value == 1]

        self.input_ids = tokenizer.convert_tokens_to_ids(self.tokens)
        self.input_mask = [1] * len(self.input_ids)

        if self.labels is not None:
          label_mapping = kwargs.get("label_mapping")
          if isinstance(self.labels, str):
            # Single Label Classfication
            self.label_ids = label_mapping[self.labels]
          elif isinstance(self.labels, list):
            # Multi Label Classification
            label_size = len(label_mapping)
            self.label_ids = [0] * label_size
            for label in self.labels:
              self.label_ids[label_mapping[label]] = 1
        else:
          self.label_ids = None

  @classmethod
  def from_structure(cls, structure):
    return cls(text=structure.text)

  @classmethod
  def from_json(cls, example):
    return cls(text=" ".join(example["tokens"]),
               labels=example.get("labels", None))

  @property
  def len(self):
    return len(self.input_ids)

class SingletonFeature(Feature):
  """Singleton Features.
  """
  def __init__(self, *inputs, **kwargs):
    super(SingletonFeature, self).__init__()
    self.input_ids = kwargs.pop("input_ids")
    self.input_mask = kwargs.pop("input_mask")
    self.segment_ids = kwargs.pop("segment_ids")
    self.is_head = kwargs.pop("is_head")
    self.label_ids = kwargs.pop("label_ids")

class SingletonMiniBatch(MiniBatch):
  """

  """
  def __init__(self, *inputs, **kwargs):
    super(SingletonMiniBatch, self).__init__(*inputs, **kwargs)

  def generate_input(self, device, use_label):
    inputs = {}
    inputs["task_name"] = self.task_name
    inputs["input_ids"] = create_tensor(self.input_features, "input_ids",
                                        torch.long, device)
    inputs["input_mask"] = create_tensor(self.input_features, "input_mask",
                                         torch.long, device)
    inputs["segment_ids"] = create_tensor(self.input_features, "segment_ids",
                                          torch.long, device)
    inputs["input_head"] = create_tensor(self.input_features, "is_head",
                                         torch.long, device)
    if use_label:
      label_ids = create_tensor(self.input_features, "label_ids",
                                torch.long, device)
      inputs["label_ids"] = label_ids
    else:
      inputs["label_ids"] = None

    # inputs["extra_args"] = {
    #   "selected_non_final_layers": [10]
    # }
    inputs["extra_args"] = {}

    if self.config.tasks[self.task_name]["selected_non_final_layers"] is not None:
      inputs["extra_args"]["selected_non_final_layers"] = self.config.tasks[self.task_name]["selected_non_final_layers"]


    return inputs

class SingletonDataFlow(DataFlow):
  def __init__(self, config, task_name, tokenizers, label_mapping):
    super(SingletonDataFlow, self).__init__(config, task_name, tokenizers, label_mapping)
    self._inv_label_mapping = {v: k for k, v in label_mapping.items()}

  @property
  def example_class(self):
    return SingletonExample

  @property
  def minibatch_class(self):
    return SingletonMiniBatch

  def process_example(self, example: SingletonExample):
    example.process(tokenizers=self.tokenizers,
                    label_mapping=self.label_mapping)

  def convert_examples_to_features(self, examples: List[SingletonExample]):
    examples: List[SingletonExample]
    features = []

    max_token_length = max([example.len for example in examples])


    for idx, example in enumerate(examples):
      padding = [0] * (max_token_length - example.len)
      input_ids = example.input_ids + padding
      input_mask = example.input_mask + padding
      segment_ids = example.segment_ids + padding
      is_head = example.is_head + [2] * (max_token_length - example.len)

      if example.label_ids is not None:
        # We assume the label length is same as sequence length
        label_ids = example.label_ids
      else:
        label_ids = None

      features.append(SingletonFeature(
        input_ids=input_ids,
        input_mask=input_mask,
        is_head=is_head,
        segment_ids=segment_ids,
        label_ids=label_ids))

    return features

  def decode_to_labels(self, preds, mb: MiniBatch):
    labels = []
    for example, pred_logits in zip(mb.examples, preds):
      pred_probs = torch.sigmoid(pred_logits).data.cpu().numpy()
      pred_label = []
      for idx, pred_prob in enumerate(pred_probs):
        if pred_prob > 0.5:
          pred_label.append(self._inv_label_mapping[idx])
      labels.append(pred_label)
    return labels






from relogic.logickit.base.constants import (
  AUXILIARY_TRAINING, ITERATIVE_TRAINING, ADVERSARIAL_TRAINING,
  TRAIN_DISCRIMINATOR, TRAIN_GENERATOR)
import numpy as np
import bisect
import json

def auxiliary_training(config, tasks):
  training_scheme = json.load(open(config.training_scheme_file))
  datasets = {task.name: task.train_set for task in tasks}
  mbs = {task.name: task.train_set.endless_minibatches(config.tasks[task.name]["train_batch_size"],
                                                       sequential=(config.tasks[task.name]["dataset_type"] == "sequential"))
          for task in tasks}
  warmup = training_scheme["warmup"]

  warmup_tasks = warmup["tasks"]
  warmup_task_weights = [np.sqrt(datasets[task_name].size) for task_name in warmup_tasks]
  warmup_id2task = {idx: task for idx, task in enumerate(warmup_tasks)}
  warmup_thresholds = np.cumsum([w / np.sum(warmup_task_weights) for w in warmup_task_weights])
  warmup_steps = warmup["steps"]
  # raise NotImplementedError()
  for i in range(warmup_steps):
    dataset_ind = bisect.bisect(warmup_thresholds, np.random.random())
    yield next(mbs[warmup_id2task[dataset_ind]])

  auxiliary = training_scheme["auxiliary"]
  auxiliary_tasks = auxiliary["tasks"]
  while True:
    for auxiliary_task in auxiliary_tasks:
      dataset_ind = bisect.bisect(warmup_thresholds, np.random.random())
      yield next(mbs[warmup_id2task[dataset_ind]])
      yield next(mbs[auxiliary_task])

def iterative_training(config, tasks):
  training_scheme = json.load(open(config.training_scheme_file))
  mbs = {task.name: task.train_set.endless_minibatches(config.tasks[task.name]["train_batch_size"],
                                                       sequential=(config.tasks[task.name]["dataset_type"] == "sequential"))
         for task in tasks}

  iterative_stages = training_scheme["iterative_stages"]
  while True:
    for stage in iterative_stages:
      stage_steps = stage["steps"]
      stage_tasks = stage["tasks"]
      for i in range(stage_steps):
        for task in stage_tasks:
          yield next(mbs[task])

def adversarial_training(config, tasks):
  training_scheme = json.load(open(config.training_scheme_file))
  # Let's assume that we only have one source_task and one target_task
  source_task = training_scheme["source_task"]
  target_task = training_scheme["target_task"]
  discriminator_training_data = {
    task.name: task.train_set.endless_minibatches(
      config.tasks[task.name]["train_batch_size"],
      sequential=(config.tasks[task.name]["dataset_type"] == "sequential")) for task in tasks}
  generator_training_data = {
    task.name: task.train_set.endless_minibatches(
      config.tasks[task.name]["train_batch_size"],
      sequential=(config.tasks[task.name]["dataset_type"] == "sequential")) for task in tasks}

  n_critic_dis = training_scheme["n_critic_dis"]
  n_critic_gen = training_scheme["n_critic_gen"]

  print("----------------")
  print("Training Scheme")
  print(training_scheme)

  while True:
    for i in range(n_critic_dis):
      yield TRAIN_DISCRIMINATOR, next(discriminator_training_data[source_task]), next(discriminator_training_data[target_task])
    for i in range(n_critic_gen):
      yield TRAIN_GENERATOR, next(generator_training_data[source_task]), next(generator_training_data[target_task])



TRAINING_SCHEME = {
  AUXILIARY_TRAINING: auxiliary_training,
  ITERATIVE_TRAINING: iterative_training,
  ADVERSARIAL_TRAINING: adversarial_training,
}


from typing import List

import torch

from relogic.logickit.dataflow import DataFlow, Example, Feature, MiniBatch
from relogic.logickit.tokenizer.tokenization import BertTokenizer
from relogic.logickit.utils import create_tensor



class ECPExample(Example):
  """EXP Example contains the attributes and functionality of
  a Emotion-Cause Pair example.

  Currently the implementation is based on Chinese.

  Args:
    text (str): A document string.
    label (List[Tuple]): A list of labels. According to the statistics
      of the data, most(89%) of the examples only contains one emotion-cause
      pair.

  """
  def __init__(self, text, clause_spans=None, labels=None):
    super(ECPExample, self).__init__()
    self.text = text
    self.clause_spans = clause_spans
    self.raw_tokens = text.split()
    self.labels = labels
    self.label_padding = '0'

  def process(self, tokenizers, *inputs, **kwargs):
    """Process the ECP example.

    This process requires the tokenizers.

    :param tokenizers:
    :param inputs:
    :param kwargs:

    """
    for tokenizer in tokenizers.values():
      if isinstance(tokenizer, BertTokenizer):
        # BERT process part
        self.text_tokens, _ = tokenizer.tokenize(self.text)

        # Some problem with Chinese Indexing
        # for token list (Chinese characters), they will be combined into a string without space.
        # Until now the spans index is still correct.
        # But after tokenization, something goes wrong. Because some numbers will be
        # merged (several digits will be merged)
        # Then the index information is wrong here.
        self.tokens = ["[CLS]"] + self.text_tokens + ["[SEP]"]
        self.char_index_to_token_index = []

        for idx, token in enumerate(self.tokens[1:]):
          # Because we use exclusive representation, so we process extra [SEP] token for padding
          self.char_index_to_token_index.extend([idx+1] * len(token))

        self.segment_ids = [0] * (len(self.text_tokens) + 2)

        self.input_ids = tokenizer.convert_tokens_to_ids(self.tokens)
        self.input_mask = [1] * len(self.input_ids)

        if self.labels is not None:
          label_mapping = kwargs.get("label_mapping")
          self.label_padding_id = label_mapping[self.label_padding]
          self.label_ids = []
          for label in self.labels:
            # Because we add [SEP] in front of each sentence, so we need to shift
            # the index to right for one step
            index_shift_right_label = tuple([self.char_index_to_token_index[item] for item in label])
            label_tuple = index_shift_right_label + (label_mapping["1"],)
            self.label_ids.append(label_tuple)

  @classmethod
  def from_structure(cls, structure):
    """Implementation of converting structure into ECPExample.
    TODO: consider how to automatically generate the clause spans"""

    return cls(text=structure.text)

  @classmethod
  def from_json(cls, example):
    """Implementation of converting json object into ECPExample.
    Currently we assume the language is Chinese. So there is
    no space between tokens. TODO: Make it general
    """
    return cls(text="".join(example["tokens"]),
               clause_spans=example.get("clause_spans", None),
               labels=example.get("labels", None))

  @property
  def len(self):
    return len(self.tokens)

  @property
  def label_len(self):
    if hasattr(self, "label_ids"):
      return len(self.label_ids)
    return 0


class ECPFeature(Feature):
  """ECP features.

  """
  def __init__(self, *inputs, **kwargs):
    super(ECPFeature, self).__init__()
    # BERT based feature
    self.input_ids = kwargs.pop("input_ids")
    self.input_mask = kwargs.pop("input_mask")
    self.segment_ids = kwargs.pop("segment_ids")
    self.label_ids = kwargs.pop("label_ids")
    self.clause_candidates = kwargs.pop("clause_candidates")


class ECPMiniBatch(MiniBatch):
  """Minibatch Implementation based on ECP task.

  """
  def __init__(self, *inputs, **kwargs):
    super(ECPMiniBatch, self).__init__(*inputs, **kwargs)

  def generate_input(self, device, use_label):
    """Generate tensors based on ECP Features"""
    # BERT based features
    inputs = {}
    inputs["task_name"] = self.task_name
    inputs["input_ids"] = create_tensor(self.input_features, "input_ids",
                                        torch.long, device)
    inputs["input_mask"] = create_tensor(self.input_features, "input_mask",
                                         torch.long, device)
    inputs["segment_ids"] = create_tensor(self.input_features, "segment_ids",
                                          torch.long, device)
    inputs["label_ids"] = create_tensor(self.input_features, "label_ids", 
                                          torch.long, device)
    inputs["clause_candidates"] = create_tensor(self.input_features, "clause_candidates",
                                          torch.long, device)
    inputs["extra_args"] = {}
    return inputs

class ECPDataFlow(DataFlow):
  """DataFlow implementation based on ECP task."""
  def __init__(self, config, task_name, tokenizers, label_mapping):
    super(ECPDataFlow, self).__init__(config, task_name, tokenizers, label_mapping)

  @property
  def example_class(self):
    return ECPExample

  @property
  def minibatch_class(self):
    return ECPMiniBatch

  def process_example(self, example: ECPExample):
    """Process ECP example"""
    example.process(tokenizers=self.tokenizers,
                    label_mapping=self.label_mapping)

  def convert_examples_to_features(self, examples: List[ECPExample]):
    examples: List[ECPExample]
    features = []

    # BERT based variables
    max_token_length = max([example.len for example in examples])
    max_clause_candidate_length = max([len(example.clause_spans) for example in examples])
    try:
      max_label_length = max([example.label_len for example in examples])
    except:
      max_label_length = None

    for idx, example in enumerate(examples):
      # BERT based feature process
      padding = [0] * (max_token_length - example.len)
      input_ids = example.input_ids + padding
      input_mask = example.input_mask + padding
      segment_ids = example.segment_ids + padding

      index_shift_right_clause_candidates = [(
        example.char_index_to_token_index[span[0]], example.char_index_to_token_index[span[1]])
        for span in example.clause_spans]
      clause_candidates = index_shift_right_clause_candidates + [
        (1, 0)
      ] * (max_clause_candidate_length - len(example.clause_spans))

      if example.label_ids is not None:
        label_ids = example.label_ids + [
          (1, 0, 1, 0, example.label_padding_id)
        ] * (max_label_length - example.label_len)
      else:
        label_ids = None

      features.append(
        ECPFeature(input_ids=input_ids,
                   input_mask=input_mask,
                   segment_ids=segment_ids,
                   label_ids=label_ids,
                   clause_candidates=clause_candidates))
    return features




from dataclasses import dataclass

@dataclass
class GeneralConfig(object):
  """Model configuration for training and test.
  
  Args:
    mode (str): The trainer mode can be one of the following choice:
      "train", "dev", "eval", "finetune"
    output_dir (str): A full directory path for saving the model configuration,
      parameters and prediction dumps.
    restore_path (str): Model configuration and parameters will be 
      restored from this full directory path.
    max_seq_length (int): Limited by the contextual model and GPU memory,
      the max sentence length is limited in this model. Default: 450.
    max_query_length (int): This is reserved for reading comprehension model.
      There are two components in reading comprehension model: query and paragraph.
      In the model, the query max length is limited by this argument. Default: 64.
    doc_stride (int): This is reserved for reading comprehension model. For long paragraph,
      it will be devided into several sub-paragraph by a sliding window whose window size
      is determined by the this argument. Default: 128.
    do_lower_case (bool): If True, convert the original text to lowercase. Default: False
    train_file (str): The file names of train file in :data:`raw_data_path`, segmented by ','. 
    dev_file (str): The file names of dev file in :data:`raw_data_path`, segmented by ','.
    test_file (str): The file names of test file in :data:`raw_data_path`, segmented by ','.
    
  """
  # IO
  mode: str
  output_dir: str
  restore_path: str

  max_seq_length: int = 450
  max_query_length: int = 64
  doc_stride: int = 128

  do_lower_case: bool = False

  train_file: str = "train.json"
  dev_file: str = "dev.json"
  test_file: str = "test.json"

  
  # Task Definition
  
  # Task Related Configuration
  
  # Hyper-parameter
  
  # Training
  
  # Semi-supervised
  
  # Training
  
  def load_from_json(self, config):
    pass
  
  def load_from_namespace(self, config):
    pass
  
  def load_from_json_file(self, config_path):
    pass
  


import json
from types import SimpleNamespace

import relogic.utils.crash_on_ipy
from relogic.logickit.base.constants import PARALLEL_MAPPING_TASK
from relogic.logickit.dataflow import TASK_TO_DATAFLOW_CLASS_MAP, ParallelDataFlow
from relogic.logickit.tokenizer.tokenization import BertTokenizer

config = SimpleNamespace(
  **{
    "buckets": [(0, 100), (100, 250), (250, 512)],
    "max_seq_length": 512
  })

tokenizers = {
  "BPE": BertTokenizer.from_pretrained("bert-base-multilingual-cased", do_lower_case=False, pretokenized=True),
}

dataflow: ParallelDataFlow = TASK_TO_DATAFLOW_CLASS_MAP[PARALLEL_MAPPING_TASK](
  task_name=PARALLEL_MAPPING_TASK,
  config=config,
  tokenizers=tokenizers,
  label_mapping=None)

examples = [{
  "text_a": "Die deutschsprachige Gemeinschaft in Belgien",
  "text_b": "Germanspeaking Community of Belgium",
  "alignment": [[0, 2, 3, 4], [0, 1, 2, 3]]
}, {
  "text_a": "The European Social Fund , created in 1957 , is the European Union ’ s main financial instrument for investing in people .",
  "text_b": "Het Europees Sociaal Fonds ( ESF ) , dat werd opgericht in 1957 , is het voornaamste nanciële instrument van de Europese Unie om te investeren in mensen .",
  "alignment": [[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 15, 17, 18, 19, 20, 21, 22],
                [0, 1, 2, 3, 7, 10, 11, 12, 13, 14, 15, 21, 22, 17, 16, 18, 23, 24, 26, 27, 28]]
},{
  "text_a": "The author of the communication is Zdeněk Kříž , a U.S. and Czech citizen , born in 1916 in Vysoké M \u200e \u200e ýto , Czech Republic , currently residing in the United States .",
  "text_b": "El autor de la comunicación es Zdenĕk Kříž , ciudadano estadounidense y checo , nacido en 1916 en Vysoké Mýto ( República Checa ) y que reside actualmente en los Estados Unidos .",
  "alignment": [[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 14, 15, 16, 17, 18, 19, 24, 26, 28, 29, 30, 31, 32, 33, 34],
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 22, 21, 27, 25, 28, 29, 31, 30, 32]]
}]

dataflow.update_with_jsons(examples)

for mb in dataflow.get_minibatches(minibatch_size=3):
  print(mb)

raise NotImplementedError("You can start to play with data")



from relogic.logickit.scripts import squad11_eval
from relogic.logickit.scripts import squad20_eval

def squad_eval(dataset):
  if dataset == "squad11":
    return squad11_eval.evaluate
  if dataset == "squad20":
    return squad20_eval.evaluate


# coding=utf-8
# Copyright 2019 The Open AI Team Authors and The HuggingFace Inc. team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tokenization classes for OpenAI GPT."""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import json
import logging
import os
import re
import sys
import unicodedata
from io import open

import sacremoses as sm

from relogic.logickit.tokenizer.tokenization_utils import PreTrainedTokenizer

logger = logging.getLogger(__name__)

VOCAB_FILES_NAMES = {
    'vocab_file': 'vocab.json',
    'merges_file': 'merges.txt',
}

PRETRAINED_VOCAB_FILES_MAP = {
    'vocab_file':
    {
        'xlm-mlm-en-2048': "https://s3.amazonaws.com/models.huggingface.co/bert/xlm-mlm-en-2048-vocab.json",
        'xlm-mlm-ende-1024': "https://s3.amazonaws.com/models.huggingface.co/bert/xlm-mlm-ende-1024-vocab.json",
        'xlm-mlm-enfr-1024': "https://s3.amazonaws.com/models.huggingface.co/bert/xlm-mlm-enfr-1024-vocab.json",
        'xlm-mlm-enro-1024': "https://s3.amazonaws.com/models.huggingface.co/bert/xlm-mlm-enro-1024-vocab.json",
        'xlm-mlm-tlm-xnli15-1024': "https://s3.amazonaws.com/models.huggingface.co/bert/xlm-mlm-tlm-xnli15-1024-vocab.json",
        'xlm-mlm-xnli15-1024': "https://s3.amazonaws.com/models.huggingface.co/bert/xlm-mlm-xnli15-1024-vocab.json",
        'xlm-clm-enfr-1024': "https://s3.amazonaws.com/models.huggingface.co/bert/xlm-clm-enfr-1024-vocab.json",
        'xlm-clm-ende-1024': "https://s3.amazonaws.com/models.huggingface.co/bert/xlm-clm-ende-1024-vocab.json",
        'xlm-mlm-17-1280': "https://s3.amazonaws.com/models.huggingface.co/bert/xlm-mlm-17-1280-vocab.json",
        'xlm-mlm-100-1280': "https://s3.amazonaws.com/models.huggingface.co/bert/xlm-mlm-100-1280-vocab.json",
    },
    'merges_file':
    {
        'xlm-mlm-en-2048': "https://s3.amazonaws.com/models.huggingface.co/bert/xlm-mlm-en-2048-merges.txt",
        'xlm-mlm-ende-1024': "https://s3.amazonaws.com/models.huggingface.co/bert/xlm-mlm-ende-1024-merges.txt",
        'xlm-mlm-enfr-1024': "https://s3.amazonaws.com/models.huggingface.co/bert/xlm-mlm-enfr-1024-merges.txt",
        'xlm-mlm-enro-1024': "https://s3.amazonaws.com/models.huggingface.co/bert/xlm-mlm-enro-1024-merges.txt",
        'xlm-mlm-tlm-xnli15-1024': "https://s3.amazonaws.com/models.huggingface.co/bert/xlm-mlm-tlm-xnli15-1024-merges.txt",
        'xlm-mlm-xnli15-1024': "https://s3.amazonaws.com/models.huggingface.co/bert/xlm-mlm-xnli15-1024-merges.txt",
        'xlm-clm-enfr-1024': "https://s3.amazonaws.com/models.huggingface.co/bert/xlm-mlm-enfr-1024-merges.txt",
        'xlm-clm-ende-1024': "https://s3.amazonaws.com/models.huggingface.co/bert/xlm-mlm-ende-1024-merges.txt",
        'xlm-mlm-17-1280': "https://s3.amazonaws.com/models.huggingface.co/bert/xlm-mlm-17-1280-merges.txt",
        'xlm-mlm-100-1280': "https://s3.amazonaws.com/models.huggingface.co/bert/xlm-mlm-100-1280-merges.txt",
    },
}

PRETRAINED_POSITIONAL_EMBEDDINGS_SIZES = {
    'xlm-mlm-en-2048': 512,
    'xlm-mlm-ende-1024': 512,
    'xlm-mlm-enfr-1024': 512,
    'xlm-mlm-enro-1024': 512,
    'xlm-mlm-tlm-xnli15-1024': 512,
    'xlm-mlm-xnli15-1024': 512,
    'xlm-clm-enfr-1024': 512,
    'xlm-clm-ende-1024': 512,
    'xlm-mlm-17-1280': 512,
    'xlm-mlm-100-1280': 512,
}

PRETRAINED_INIT_CONFIGURATION = {
    'xlm-mlm-en-2048': {"do_lowercase_and_remove_accent": True},
    'xlm-mlm-ende-1024': { "do_lowercase_and_remove_accent": True,
                            "id2lang": { "0": "de",
                                        "1": "en"},
                           "lang2id": { "de": 0,
                                        "en": 1 }},
    'xlm-mlm-enfr-1024': { "do_lowercase_and_remove_accent": True,
                           "id2lang": { "0": "en",
                                        "1": "fr"},
                           "lang2id": { "en": 0,
                                        "fr": 1 }},
    'xlm-mlm-enro-1024': { "do_lowercase_and_remove_accent": True,
                           "id2lang": { "0": "en",
                                        "1": "ro"},
                           "lang2id": { "en": 0,
                                        "ro": 1 }},
    'xlm-mlm-tlm-xnli15-1024': { "do_lowercase_and_remove_accent": True,
                                 "id2lang": {   "0": "ar",
                                                "1": "bg",
                                                "2": "de",
                                                "3": "el",
                                                "4": "en",
                                                "5": "es",
                                                "6": "fr",
                                                "7": "hi",
                                                "8": "ru",
                                                "9": "sw",
                                                "10": "th",
                                                "11": "tr",
                                                "12": "ur",
                                                "13": "vi",
                                                "14": "zh"},
                                 "lang2id": {   "ar": 0,
                                                "bg": 1,
                                                "de": 2,
                                                "el": 3,
                                                "en": 4,
                                                "es": 5,
                                                "fr": 6,
                                                "hi": 7,
                                                "ru": 8,
                                                "sw": 9,
                                                "th": 10,
                                                "tr": 11,
                                                "ur": 12,
                                                "vi": 13,
                                                "zh": 14 }},
    'xlm-mlm-xnli15-1024': { "do_lowercase_and_remove_accent": True,
                             "id2lang": {   "0": "ar",
                                                "1": "bg",
                                                "2": "de",
                                                "3": "el",
                                                "4": "en",
                                                "5": "es",
                                                "6": "fr",
                                                "7": "hi",
                                                "8": "ru",
                                                "9": "sw",
                                                "10": "th",
                                                "11": "tr",
                                                "12": "ur",
                                                "13": "vi",
                                                "14": "zh"},
                                 "lang2id": {   "ar": 0,
                                                "bg": 1,
                                                "de": 2,
                                                "el": 3,
                                                "en": 4,
                                                "es": 5,
                                                "fr": 6,
                                                "hi": 7,
                                                "ru": 8,
                                                "sw": 9,
                                                "th": 10,
                                                "tr": 11,
                                                "ur": 12,
                                                "vi": 13,
                                                "zh": 14 }},
    'xlm-clm-enfr-1024': { "do_lowercase_and_remove_accent": True,
                           "id2lang": { "0": "en",
                                        "1": "fr"},
                           "lang2id": { "en": 0,
                                        "fr": 1 }},
    'xlm-clm-ende-1024': { "do_lowercase_and_remove_accent": True,
                           "id2lang": { "0": "de",
                                        "1": "en"},
                           "lang2id": { "de": 0,
                                        "en": 1 }},
    'xlm-mlm-17-1280': {"do_lowercase_and_remove_accent": False,
                        "id2lang": {
                            "0": "ar",
                            "1": "de",
                            "2": "en",
                            "3": "es",
                            "4": "fr",
                            "5": "hi",
                            "6": "it",
                            "7": "ja",
                            "8": "ko",
                            "9": "nl",
                            "10": "pl",
                            "11": "pt",
                            "12": "ru",
                            "13": "sv",
                            "14": "tr",
                            "15": "vi",
                            "16": "zh"
                        },
                        "lang2id": {
                            "ar": 0,
                            "de": 1,
                            "en": 2,
                            "es": 3,
                            "fr": 4,
                            "hi": 5,
                            "it": 6,
                            "ja": 7,
                            "ko": 8,
                            "nl": 9,
                            "pl": 10,
                            "pt": 11,
                            "ru": 12,
                            "sv": 13,
                            "tr": 14,
                            "vi": 15,
                            "zh": 16}},
    'xlm-mlm-100-1280': {"do_lowercase_and_remove_accent": False,
                        "id2lang": {
                            "0": "af",
                            "1": "als",
                            "2": "am",
                            "3": "an",
                            "4": "ang",
                            "5": "ar",
                            "6": "arz",
                            "7": "ast",
                            "8": "az",
                            "9": "bar",
                            "10": "be",
                            "11": "bg",
                            "12": "bn",
                            "13": "br",
                            "14": "bs",
                            "15": "ca",
                            "16": "ceb",
                            "17": "ckb",
                            "18": "cs",
                            "19": "cy",
                            "20": "da",
                            "21": "de",
                            "22": "el",
                            "23": "en",
                            "24": "eo",
                            "25": "es",
                            "26": "et",
                            "27": "eu",
                            "28": "fa",
                            "29": "fi",
                            "30": "fr",
                            "31": "fy",
                            "32": "ga",
                            "33": "gan",
                            "34": "gl",
                            "35": "gu",
                            "36": "he",
                            "37": "hi",
                            "38": "hr",
                            "39": "hu",
                            "40": "hy",
                            "41": "ia",
                            "42": "id",
                            "43": "is",
                            "44": "it",
                            "45": "ja",
                            "46": "jv",
                            "47": "ka",
                            "48": "kk",
                            "49": "kn",
                            "50": "ko",
                            "51": "ku",
                            "52": "la",
                            "53": "lb",
                            "54": "lt",
                            "55": "lv",
                            "56": "mk",
                            "57": "ml",
                            "58": "mn",
                            "59": "mr",
                            "60": "ms",
                            "61": "my",
                            "62": "nds",
                            "63": "ne",
                            "64": "nl",
                            "65": "nn",
                            "66": "no",
                            "67": "oc",
                            "68": "pl",
                            "69": "pt",
                            "70": "ro",
                            "71": "ru",
                            "72": "scn",
                            "73": "sco",
                            "74": "sh",
                            "75": "si",
                            "76": "simple",
                            "77": "sk",
                            "78": "sl",
                            "79": "sq",
                            "80": "sr",
                            "81": "sv",
                            "82": "sw",
                            "83": "ta",
                            "84": "te",
                            "85": "th",
                            "86": "tl",
                            "87": "tr",
                            "88": "tt",
                            "89": "uk",
                            "90": "ur",
                            "91": "uz",
                            "92": "vi",
                            "93": "war",
                            "94": "wuu",
                            "95": "yi",
                            "96": "zh",
                            "97": "zh_classical",
                            "98": "zh_min_nan",
                            "99": "zh_yue"
                        },
                        "lang2id": {
                            "af": 0,
                            "als": 1,
                            "am": 2,
                            "an": 3,
                            "ang": 4,
                            "ar": 5,
                            "arz": 6,
                            "ast": 7,
                            "az": 8,
                            "bar": 9,
                            "be": 10,
                            "bg": 11,
                            "bn": 12,
                            "br": 13,
                            "bs": 14,
                            "ca": 15,
                            "ceb": 16,
                            "ckb": 17,
                            "cs": 18,
                            "cy": 19,
                            "da": 20,
                            "de": 21,
                            "el": 22,
                            "en": 23,
                            "eo": 24,
                            "es": 25,
                            "et": 26,
                            "eu": 27,
                            "fa": 28,
                            "fi": 29,
                            "fr": 30,
                            "fy": 31,
                            "ga": 32,
                            "gan": 33,
                            "gl": 34,
                            "gu": 35,
                            "he": 36,
                            "hi": 37,
                            "hr": 38,
                            "hu": 39,
                            "hy": 40,
                            "ia": 41,
                            "id": 42,
                            "is": 43,
                            "it": 44,
                            "ja": 45,
                            "jv": 46,
                            "ka": 47,
                            "kk": 48,
                            "kn": 49,
                            "ko": 50,
                            "ku": 51,
                            "la": 52,
                            "lb": 53,
                            "lt": 54,
                            "lv": 55,
                            "mk": 56,
                            "ml": 57,
                            "mn": 58,
                            "mr": 59,
                            "ms": 60,
                            "my": 61,
                            "nds": 62,
                            "ne": 63,
                            "nl": 64,
                            "nn": 65,
                            "no": 66,
                            "oc": 67,
                            "pl": 68,
                            "pt": 69,
                            "ro": 70,
                            "ru": 71,
                            "scn": 72,
                            "sco": 73,
                            "sh": 74,
                            "si": 75,
                            "simple": 76,
                            "sk": 77,
                            "sl": 78,
                            "sq": 79,
                            "sr": 80,
                            "sv": 81,
                            "sw": 82,
                            "ta": 83,
                            "te": 84,
                            "th": 85,
                            "tl": 86,
                            "tr": 87,
                            "tt": 88,
                            "uk": 89,
                            "ur": 90,
                            "uz": 91,
                            "vi": 92,
                            "war": 93,
                            "wuu": 94,
                            "yi": 95,
                            "zh": 96,
                            "zh_classical": 97,
                            "zh_min_nan": 98,
                            "zh_yue": 99
                        }},
}

def get_pairs(word):
    """
    Return set of symbol pairs in a word.
    word is represented as tuple of symbols (symbols being variable-length strings)
    """
    pairs = set()
    prev_char = word[0]
    for char in word[1:]:
        pairs.add((prev_char, char))
        prev_char = char
    return pairs


def lowercase_and_remove_accent(text):
    """
    Lowercase and strips accents from a piece of text based on
    https://github.com/facebookresearch/XLM/blob/master/tools/lowercase_and_remove_accent.py
    """
    text = ' '.join(text)
    text = text.lower()
    text = unicodedata.normalize("NFD", text)
    output = []
    for char in text:
        cat = unicodedata.category(char)
        if cat == "Mn":
            continue
        output.append(char)
    return "".join(output).lower().split(' ')


def replace_unicode_punct(text):
    '''
    Port of https://github.com/moses-smt/mosesdecoder/blob/master/scripts/tokenizer/replace-unicode-punctuation.perl
    '''
    text = text.replace('，', ',')
    text = re.sub(r'。\s*', '. ', text)
    text = text.replace('、', ',')
    text = text.replace('”', '"')
    text = text.replace('“', '"')
    text = text.replace('∶', ':')
    text = text.replace('：', ':')
    text = text.replace('？', '?')
    text = text.replace('《', '"')
    text = text.replace('》', '"')
    text = text.replace('）', ')')
    text = text.replace('！', '!')
    text = text.replace('（', '(')
    text = text.replace('；', ';')
    text = text.replace('１', '"')
    text = text.replace('」', '"')
    text = text.replace('「', '"')
    text = text.replace('０', '0')
    text = text.replace('３', '3')
    text = text.replace('２', '2')
    text = text.replace('５', '5')
    text = text.replace('６', '6')
    text = text.replace('９', '9')
    text = text.replace('７', '7')
    text = text.replace('８', '8')
    text = text.replace('４', '4')
    text = re.sub(r'．\s*', '. ', text)
    text = text.replace('～', '~')
    text = text.replace('’', '\'')
    text = text.replace('…', '...')
    text = text.replace('━', '-')
    text = text.replace('〈', '<')
    text = text.replace('〉', '>')
    text = text.replace('【', '[')
    text = text.replace('】', ']')
    text = text.replace('％', '%')
    return text


def remove_non_printing_char(text):
    '''
    Port of https://github.com/moses-smt/mosesdecoder/blob/master/scripts/tokenizer/remove-non-printing-char.perl
    '''
    output = []
    for char in text:
        cat = unicodedata.category(char)
        if cat.startswith('C'):
            continue
        output.append(char)
    return "".join(output)


def romanian_preprocessing(text):
    '''Sennrich's WMT16 scripts for Romanian preprocessing, used by model `xlm-mlm-enro-1024`'''
    # https://github.com/rsennrich/wmt16-scripts/blob/master/preprocess/normalise-romanian.py
    text = text.replace("\u015e", "\u0218").replace("\u015f", "\u0219")
    text = text.replace("\u0162", "\u021a").replace("\u0163", "\u021b")
    # https://github.com/rsennrich/wmt16-scripts/blob/master/preprocess/remove-diacritics.py
    text = text.replace("\u0218", "S").replace("\u0219", "s") #s-comma
    text = text.replace("\u021a", "T").replace("\u021b", "t") #t-comma
    text = text.replace("\u0102", "A").replace("\u0103", "a")
    text = text.replace("\u00C2", "A").replace("\u00E2", "a")
    text = text.replace("\u00CE", "I").replace("\u00EE", "i")
    return text


class XLMTokenizer(PreTrainedTokenizer):
    """
    BPE tokenizer for XLM

        - Moses preprocessing & tokenization for most supported languages

        - Language specific tokenization for Chinese (Jieba), Japanese (KyTea) and Thai (PyThaiNLP)

        - (optionally) lower case & normalize all inputs text

        - argument ``special_tokens`` and function ``set_special_tokens``, can be used to add additional symbols \
        (ex: "__classify__") to a vocabulary
        
        - `lang2id` attribute maps the languages supported by the model with their ids if provided (automatically set for pretrained vocabularies)

        - `id2lang` attributes does reverse mapping if provided (automatically set for pretrained vocabularies)

        - `do_lowercase_and_remove_accent` controle lower casing and accent (automatically set for pretrained vocabularies)
    """
    vocab_files_names = VOCAB_FILES_NAMES
    pretrained_vocab_files_map = PRETRAINED_VOCAB_FILES_MAP
    pretrained_init_configuration = PRETRAINED_INIT_CONFIGURATION
    max_model_input_sizes = PRETRAINED_POSITIONAL_EMBEDDINGS_SIZES

    def __init__(self, vocab_file, merges_file, unk_token="<unk>", bos_token="<s>",
                 sep_token="</s>", pad_token="<pad>", cls_token="</s>",
                 mask_token="<special1>", additional_special_tokens=["<special0>",
                 "<special1>", "<special2>", "<special3>", "<special4>", "<special5>",
                 "<special6>", "<special7>", "<special8>", "<special9>"],
                 lang2id=None, id2lang=None, do_lowercase_and_remove_accent=True,
                 **kwargs):
        super(XLMTokenizer, self).__init__(unk_token=unk_token, bos_token=bos_token,
                                           sep_token=sep_token, pad_token=pad_token,
                                           cls_token=cls_token, mask_token=mask_token,
                                           additional_special_tokens=additional_special_tokens,
                                           **kwargs)

        # cache of sm.MosesPunctNormalizer instance
        self.cache_moses_punct_normalizer = dict()
        # cache of sm.MosesTokenizer instance
        self.cache_moses_tokenizer = dict()
        self.lang_with_custom_tokenizer = set(['zh', 'th', 'ja'])
        # True for current supported model (v1.2.0), False for XLM-17 & 100
        self.do_lowercase_and_remove_accent = do_lowercase_and_remove_accent
        self.lang2id = lang2id
        self.id2lang = id2lang
        if lang2id is not None and id2lang is not None:
            assert len(lang2id) == len(id2lang)

        self.ja_word_tokenizer = None
        self.zh_word_tokenizer = None

        self.encoder = json.load(open(vocab_file, encoding="utf-8"))
        self.decoder = {v:k for k,v in self.encoder.items()}
        merges = open(merges_file, encoding='utf-8').read().split('\n')[:-1]
        merges = [tuple(merge.split()[:2]) for merge in merges]
        self.bpe_ranks = dict(zip(merges, range(len(merges))))
        self.cache = {}

    def moses_punct_norm(self, text, lang):
        if lang not in self.cache_moses_punct_normalizer:
            punct_normalizer = sm.MosesPunctNormalizer(lang=lang)
            self.cache_moses_punct_normalizer[lang] = punct_normalizer
        else:
            punct_normalizer = self.cache_moses_punct_normalizer[lang]
        return punct_normalizer.normalize(text)

    def moses_tokenize(self, text, lang):
        if lang not in self.cache_moses_tokenizer:
            moses_tokenizer = sm.MosesTokenizer(lang=lang)
            self.cache_moses_tokenizer[lang] = moses_tokenizer
        else:
            moses_tokenizer = self.cache_moses_tokenizer[lang]
        return moses_tokenizer.tokenize(text, return_str=False, escape=False)

    def moses_pipeline(self, text, lang):
        text = replace_unicode_punct(text)
        text = self.moses_punct_norm(text, lang)
        text = remove_non_printing_char(text)
        return text

    def ja_tokenize(self, text):
        if self.ja_word_tokenizer is None:
            try:
                import Mykytea
                self.ja_word_tokenizer = Mykytea.Mykytea('-model %s/local/share/kytea/model.bin' % os.path.expanduser('~'))
            except (AttributeError, ImportError) as e:
                logger.error("Make sure you install KyTea (https://github.com/neubig/kytea) and it's python wrapper (https://github.com/chezou/Mykytea-python) with the following steps")
                logger.error("1. git clone git@github.com:neubig/kytea.git && cd kytea")
                logger.error("2. autoreconf -i")
                logger.error("3. ./configure --prefix=$HOME/local")
                logger.error("4. make && make install")
                logger.error("5. pip install kytea")
                raise e
        return list(self.ja_word_tokenizer.getWS(text))

    @property
    def vocab_size(self):
        return len(self.encoder)

    def bpe(self, token):
        word = tuple(token[:-1]) + (token[-1] + '</w>',)
        if token in self.cache:
            return self.cache[token]
        pairs = get_pairs(word)

        if not pairs:
            return token+'</w>'

        while True:
            bigram = min(pairs, key=lambda pair: self.bpe_ranks.get(pair, float('inf')))
            if bigram not in self.bpe_ranks:
                break
            first, second = bigram
            new_word = []
            i = 0
            while i < len(word):
                try:
                    j = word.index(first, i)
                    new_word.extend(word[i:j])
                    i = j
                except:
                    new_word.extend(word[i:])
                    break

                if word[i] == first and i < len(word)-1 and word[i+1] == second:
                    new_word.append(first+second)
                    i += 2
                else:
                    new_word.append(word[i])
                    i += 1
            new_word = tuple(new_word)
            word = new_word
            if len(word) == 1:
                break
            else:
                pairs = get_pairs(word)
        word = ' '.join(word)
        if word == '\n  </w>':
            word = '\n</w>'
        self.cache[token] = word
        return word

    def _tokenize(self, text, lang='en', bypass_tokenizer=False):
        """
        Tokenize a string given language code. For Chinese, Japanese and Thai, we use a language specific tokenizerself. Otherwise, we use Moses.

        Details of tokenization:
        - [sacremoses](https://github.com/alvations/sacremoses): port of Moses
            - Install with `pip install sacremoses`
        - [pythainlp](https://github.com/PyThaiNLP/pythainlp): Thai tokenizer
            - Install with `pip install pythainlp`
        - [kytea](https://github.com/chezou/Mykytea-python): Japanese tokenizer, wrapper of [KyTea](https://github.com/neubig/kytea)
            - Install with the following steps:
            ```
            git clone git@github.com:neubig/kytea.git && cd kytea
            autoreconf -i
            ./configure --prefix=$HOME/local
            make && make install
            pip install kytea
            ```
        - [jieba](https://github.com/fxsjy/jieba): Chinese tokenizer *
            - Install with `pip install jieba`

        \* The original XLM used [Stanford Segmenter](https://nlp.stanford.edu/software/stanford-segmenter-2018-10-16.zip).
        However, the wrapper (`nltk.tokenize.stanford_segmenter`) is slow due to JVM overhead, and it will be deprecated.
        Jieba is a lot faster and pip-installable. Note there is some mismatch with the Stanford Segmenter. It should be fine
        if you fine-tune the model with Chinese supervisionself. If you want the same exact behaviour, use the original XLM
        [preprocessing script](https://github.com/facebookresearch/XLM/tree/master/tools) to tokenize the sentence externally,
        and set `bypass_tokenizer=True` to bypass the tokenizer.

        Args:
            - lang: ISO language code (default = 'en') (string). Languages should belong of the model supported languages. However, we don't enforce it.
            - bypass_tokenizer: Allow users to preprocess and tokenize the sentences externally (default = False)  (bool). If True, we only apply BPE.

        Returns:
            List of tokens.
        """
        if lang and self.lang2id and lang not in self.lang2id:
            logger.error("Supplied language code not found in lang2id mapping. Please check that your language is supported by the loaded pretrained model.")
        if bypass_tokenizer:
            text = text.split()
        elif lang not in self.lang_with_custom_tokenizer:
            text = self.moses_pipeline(text, lang=lang)
            # TODO: make sure we are using `xlm-mlm-enro-1024`, since XLM-100 doesn't have this step
            if lang == 'ro':
                text = romanian_preprocessing(text)
            text = self.moses_tokenize(text, lang=lang)
        elif lang == 'th':
            text = self.moses_pipeline(text, lang=lang)
            try:
                if 'pythainlp' not in sys.modules:
                    from pythainlp.tokenize import word_tokenize as th_word_tokenize
                else:
                    th_word_tokenize = sys.modules['pythainlp'].word_tokenize
            except (AttributeError, ImportError) as e:
                logger.error("Make sure you install PyThaiNLP (https://github.com/PyThaiNLP/pythainlp) with the following steps")
                logger.error("1. pip install pythainlp")
                raise e
            text = th_word_tokenize(text)
        elif lang == 'zh':
            try:
                if 'jieba' not in sys.modules:
                    import jieba
                else:
                    jieba = sys.modules['jieba']
            except (AttributeError, ImportError) as e:
                logger.error("Make sure you install Jieba (https://github.com/fxsjy/jieba) with the following steps")
                logger.error("1. pip install jieba")
                raise e
            text = ' '.join(jieba.cut(text))
            text = self.moses_pipeline(text, lang=lang)
            text = text.split()
        elif lang == 'ja':
            text = self.moses_pipeline(text, lang=lang)
            text = self.ja_tokenize(text)
        else:
            raise ValueError('It should not reach here')

        if self.do_lowercase_and_remove_accent and not bypass_tokenizer:
            text = lowercase_and_remove_accent(text)

        split_tokens = []
        for token in text:
            if token:
                split_tokens.extend([t for t in self.bpe(token).split(' ')])

        return split_tokens

    def _convert_token_to_id(self, token):
        """ Converts a token (str/unicode) in an id using the vocab. """
        return self.encoder.get(token, self.encoder.get(self.unk_token))

    def _convert_id_to_token(self, index):
        """Converts an index (integer) in a token (string/unicode) using the vocab."""
        return self.decoder.get(index, self.unk_token)

    def convert_tokens_to_string(self, tokens):
        """ Converts a sequence of tokens (string) in a single string. """
        out_string = ''.join(tokens).replace('</w>', ' ').strip()
        return out_string

    def build_inputs_with_special_tokens(self, token_ids_0, token_ids_1=None):
        """
        Build model inputs from a sequence or a pair of sequence for sequence classification tasks
        by concatenating and adding special tokens.
        A RoBERTa sequence has the following format:
            single sequence: <s> X </s>
            pair of sequences: <s> A </s></s> B </s>
        """
        if token_ids_1 is None:
            return [self.cls_token_id] + token_ids_0 + [self.sep_token_id]
        sep = [self.sep_token_id]
        cls = [self.cls_token_id]
        return cls + token_ids_0 + sep + token_ids_1 + sep

    def get_special_tokens_mask(self, token_ids_0, token_ids_1=None, already_has_special_tokens=False):
        """
        Retrieves sequence ids from a token list that has no special tokens added. This method is called when adding
        special tokens using the tokenizer ``prepare_for_model`` or ``encode_plus`` methods.

        Args:
            token_ids_0: list of ids (must not contain special tokens)
            token_ids_1: Optional list of ids (must not contain special tokens), necessary when fetching sequence ids
                for sequence pairs
            already_has_special_tokens: (default False) Set to True if the token list is already formated with
                special tokens for the model

        Returns:
            A list of integers in the range [0, 1]: 0 for a special token, 1 for a sequence token.
        """

        if already_has_special_tokens:
            if token_ids_1 is not None:
                raise ValueError("You should not supply a second sequence if the provided sequence of "
                                 "ids is already formated with special tokens for the model.")
            return list(map(lambda x: 1 if x in [self.sep_token_id, self.cls_token_id] else 0, token_ids_0))

        if token_ids_1 is not None:
            return [1] + ([0] * len(token_ids_0)) + [1] + ([0] * len(token_ids_1)) + [1]
        return [1] + ([0] * len(token_ids_0)) + [1]

    def create_token_type_ids_from_sequences(self, token_ids_0, token_ids_1=None):
        """
        Creates a mask from the two sequences passed to be used in a sequence-pair classification task.
        An XLM sequence pair mask has the following format:
        0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1
        | first sequence    | second sequence

        if token_ids_1 is None, only returns the first portion of the mask (0's).
        """
        sep = [self.sep_token_id]
        cls = [self.cls_token_id]
        if token_ids_1 is None:
            return len(cls + token_ids_0 + sep) * [0]
        return len(cls + token_ids_0 + sep) * [0] + len(token_ids_1 + sep) * [1]

    def save_vocabulary(self, save_directory):
        """Save the tokenizer vocabulary and merge files to a directory."""
        if not os.path.isdir(save_directory):
            logger.error("Vocabulary path ({}) should be a directory".format(save_directory))
            return
        vocab_file = os.path.join(save_directory, VOCAB_FILES_NAMES['vocab_file'])
        merge_file = os.path.join(save_directory, VOCAB_FILES_NAMES['merges_file'])

        with open(vocab_file, 'w', encoding='utf-8') as f:
            f.write(json.dumps(self.encoder, ensure_ascii=False))

        index = 0
        with open(merge_file, "w", encoding="utf-8") as writer:
            for bpe_tokens, token_index in sorted(self.bpe_ranks.items(), key=lambda kv: kv[1]):
                if index != token_index:
                    logger.warning("Saving vocabulary to {}: BPE merge indices are not consecutive."
                                   " Please check that the tokenizer is not corrupted!".format(merge_file))
                    index = token_index
                writer.write(' '.join(bpe_tokens) + u'\n')
                index += 1

        return vocab_file, merge_file



import logging
import collections
from relogic.logickit.tokenizer.tokenization import BasicTokenizer, load_vocab

from relogic.utils.file_utils import cached_path

logger = logging.getLogger(__name__)

PRETRAINED_VECTOR_ARCHIVE_MAP = {
  'wiki-news-300d-1M': "https://github.com/Impavidity/relogic/raw/master/relogic/logickit/vocabs/wiki-news-300d-1M.txt",
  'aligned-fasttext-ar': "https://github.com/Impavidity/relogic/raw/master/relogic/logickit/vocabs/aligned-fasttext-ar.txt",
  'aligned-fasttext-bn': "https://github.com/Impavidity/relogic/raw/master/relogic/logickit/vocabs/aligned-fasttext-bn.txt",
  'aligned-fasttext-fr': "https://github.com/Impavidity/relogic/raw/master/relogic/logickit/vocabs/aligned-fasttext-fr.txt",
  'aligned-fasttext-hi': "https://github.com/Impavidity/relogic/raw/master/relogic/logickit/vocabs/aligned-fasttext-hi.txt",
  'aligned-fasttext-zh': "https://github.com/Impavidity/relogic/raw/master/relogic/logickit/vocabs/aligned-fasttext-zh.txt"
}

class FasttextTokenizer(object):
  """

  """

  def __init__(self, vocab_file, do_lower_case=True):
    self.vocab = load_vocab(vocab_file)

    self.ids_to_tokens = collections.OrderedDict(
      [(ids, tok) for tok, ids in self.vocab.items()])
    self.basic_tokenizer = BasicTokenizer(do_lower_case=do_lower_case)


  def tokenize(self, text):
    # tokens, _ = self.basic_tokenizer.tokenize(text)
    tokens = text.split(" ")
    return tokens

  def convert_tokens_to_ids(self, tokens):
    """Converts a sequence of tokens into ids using the vocab."""
    ids = []
    for token in tokens:
      ids.append(self.vocab.get(token, 0))
    return ids

  def convert_ids_to_tokens(self, ids):
    """Converts a sequence of ids in fasttext tokens using the vocab."""
    tokens = []
    for i in ids:
      tokens.append(self.ids_to_tokens[i])
    return tokens


  @classmethod
  def from_pretrained(cls, pretrained_model_name_or_path, cache_dir=None, *inputs, **kwargs):
    if pretrained_model_name_or_path in PRETRAINED_VECTOR_ARCHIVE_MAP:
      vocab_file = PRETRAINED_VECTOR_ARCHIVE_MAP[pretrained_model_name_or_path]
    else:
      vocab_file = pretrained_model_name_or_path
    try:
      resolved_vocab_file = cached_path(vocab_file, cache_dir=cache_dir)
    except EnvironmentError:
      logger.error(
        "Model name '{}' was not found in model name list ({}). "
        "We assumed '{}' was a path or url but couldn't find any file "
        "associated to this path or url.".format(
          pretrained_model_name_or_path,
          ', '.join(PRETRAINED_VECTOR_ARCHIVE_MAP.keys()),
          vocab_file))
      return None
    if resolved_vocab_file == vocab_file:
      logger.info("loading vocabulary file {}".format(vocab_file))
    else:
      logger.info("loading vocabulary file {} from cache at {}".format(
        vocab_file, resolved_vocab_file))
    tokenizer = cls(resolved_vocab_file, *inputs, **kwargs)
    return tokenizer



from dataclasses import dataclass, field
import regex as re
from typing import List, Union

from relogic.structures.structure import Structure
from relogic.structures.token import Token
from relogic.structures.span import Span
from transformers.tokenization_bert import BasicTokenizer

basic_tokenizer = BasicTokenizer(do_lower_case=False)

PAT = re.compile(r"""'s|'t|'re|'ve|'m|'ll|'d| ?\p{L}+| ?\p{N}+| ?[^\s\p{L}\p{N}]+|\s+(?!\S)|\s+""")

@dataclass
class Sentence(Structure):
  idx : int = None
  text: str = None
  tokens: List[Token] = field(default_factory=list)
  text_: str = None
  pos: List = field(default_factory=list)
  spans: List[Span] = field(default_factory=list)
  predicate_text: str = None
  predicate_index: int = None
  predicates: List = field(default_factory=list)
  srl_labels: List = field(default_factory=list)

  tokenizer: str = "space"

  def __post_init__(self):
    if self.text:
      if self.tokenizer == "space":
        self.tokens = [Token(token) for token in self.text.split(" ")]
      elif self.tokenizer == "basic":
        self.tokens = [Token(token) for token in basic_tokenizer.tokenize(self.text)]
      elif self.tokenizer == "gpt2":
        self.tokens = [Token(token.strip(" ")) for token in re.findall(PAT, self.text)]
      else:
        raise ValueError("Unknown tokenizer method {}".format(self.tokenizer))
    # if self.text:
    #   # Currently we only use split by tokens
    #   self.tokens = [Token(token) for token in self.text.split()]

  def add_token(self, token: Union[str, Token]):
    """Append a token into tokens.
    
    Args:
      token (str, Token): 
    """
    if isinstance(token, str):
      self.tokens.append(Token(text=token))
    else:
      self.tokens.append(token)

  def add_span(self, span: Span):
    self.spans.append(span)

  @property
  def length(self):
    return len(self.tokens)

  @property
  def tokenized_text(self):
    if self.text_ is None:
      # if len(self.tokens) == 0:
        # raise ValueError("The sentence {} is not tokenized.".format(self.text))

      self.text_ = " ".join([token.text for token in self.tokens])
    return self.text_

  def convert_to_json(self):
    return {
      "text": self.tokenized_text,
      "predicates": self.predicates,
      "srl_labels": self.srl_labels
    }

  def __getitem__(self, item):
    return self.tokens[item]



from relogic.logickit.modules.span_extractors.endpoint_span_extractor import EndpointSpanExtractor
from relogic.logickit.modules.span_extractors.self_attentive_span_extractor import SelfAttentiveSpanExtractor
from relogic.logickit.modules.span_extractors.attentive_span_extractor import AttentiveSpanExtractor


import torch.nn as nn
import torch.nn.functional as F
import torch
from relogic.logickit.base import utils
from relogic.logickit.utils import utils

class GenRepr(nn.Module):
  def __init__(self, config, task_name, n_classes=None):
    super(GenRepr, self).__init__()
    self.config = config
    self.task_name = task_name

  def forward(self, *inputs, **kwargs):
    encoding_results = kwargs.pop("encoding_results", None)
    if encoding_results is not None and "selected_non_final_layers_features" in encoding_results:
      features = encoding_results["selected_non_final_layers_features"][0]
      # We assume only one layer for now
    else:
      features = kwargs.pop("features")

    text_mask = (kwargs["input_mask"]).float()

    features_sum = torch.sum(features * text_mask.unsqueeze(-1), -2)
    length = torch.sum(kwargs["input_mask"] > 0, dim=-1)
    features_avg = features_sum / length.unsqueeze(-1).float()

    return features_avg


import abc
from relogic.logickit.scorer.scorer import Scorer

class WordLevelScorer(Scorer, metaclass=abc.ABCMeta):
  def __init__(self):
    super(WordLevelScorer, self).__init__()
    self._total_loss = 0
    self._total_words = 0
    self._examples = []
    self._preds = []

  def update(self, mbs, predictions, loss, extra):
    # TODO: need to double check
    # TODO migrate interface!!!
    predictions = predictions["logits"]
    super(WordLevelScorer, self).update(mbs, predictions, loss, extra)
    n_words = 0
    for example, preds in zip(mbs.examples, predictions):
      self._examples.append(example)
      self._preds.append(preds)
      n_words += example.bucketing_len
    self._total_loss += loss * n_words
    self._total_words += n_words

  def get_loss(self):
    return self._total_loss / max(1, self._total_words)


from .fasttext_tokenization import FasttextTokenizer
from .tokenizer_roberta_xlm import RobertaXLMTokenizer
from .tokenization import BertTokenizer as CustomizedBertTokenizer
from transformers.tokenization_bert import BertTokenizer

NAME_TO_TOKENIZER_MAP = {
  "xlmr": RobertaXLMTokenizer,
  "fasttext": FasttextTokenizer,
  "customized_bert": CustomizedBertTokenizer,
  "bert": BertTokenizer
}


import json
from types import SimpleNamespace

import relogic.utils.crash_on_ipy
from relogic.logickit.base.constants import MIXSENT_TASK
from relogic.logickit.dataflow import TASK_TO_DATAFLOW_CLASS_MAP, MixSentDataFlow
from relogic.logickit.tokenizer.tokenizer_roberta_xlm import RobertaXLMTokenizer

config = SimpleNamespace(
  **{
    "buckets": [(0, 100), (100, 250), (250, 512)],
    "max_seq_length": 512
  })

tokenizers = {
  "xlmr": RobertaXLMTokenizer.from_pretrained("xlmr.large.v0")
}

dataflow: MixSentDataFlow = TASK_TO_DATAFLOW_CLASS_MAP[MIXSENT_TASK](
  task_name=MIXSENT_TASK,
  config=config,
  tokenizers=tokenizers,
  label_mapping=None)

examples = [
  {"text_a": ["EFE", "-", "Cantabria", "Madrid", ",", "23", "may", "(", "EFE", ")", "."],
   "text_b": ["De", "este", "modo", ",", "podr\u00edan", "ser", "unos", "1.500", "los", "milicianos",
              "que", "se", "han", "rendido", "en", "los", "tres", "\u00faltimos", "d\u00edas", ",",
              "mientras", "se", "supone", "que", "los", "mil", "componentes", "restantes", "del", "ESL",
              "tratan", "de", "llegar", "a", "Israel", "."],
   "text_c": ["-", "Cantabria", "Madrid", ",", "23", "d\u00edas", ",", "mientras", "se", "supone", "que",
              "los", "mil", "componentes", "restantes", "del", "ESL", "tratan", "de", "llegar", "a", "Israel"],
   "span_a": [1, 6], "span_b": [18, 35], "span_c_a": [0, 5], "span_c_b": [5, 22]},
{"text_a": ["El", "entrenador", "de", "la", "Real", "Sociedad", "Javier", "Clemente", "dijo", "hoy", "que", "el",
            "club", "guipuzcoano", "no", "tiene", "opciones", "de", "acceder", "al", "mercado", "espa\u00f1ol",
            ",", "ya", "que", "a", "su", "juicio", "\"", "es", "imposible", "\"", "fichar", "jugadores", "de", "\"",
            "cierto", "nivel", "\"", ",", "por", "lo", "que", "afirm\u00f3", "que", "esa", "idea", "\"", "hay", "que",
            "quit\u00e1rsela", "de", "la", "cabeza", "\"", "."],
 "text_b": ["Este", "portavoz", "sindical", "dijo", "tras", "la", "desconvocatoria", "de", "la", "huelga", "que",
            "este", "preacuerdo", "\"", "puede", "ser", "perjudicial", "y", "regresivo", "para", "las", "condiciones",
            "laborales", "\"", "de", "los", "trabajadores", ",", "si", "bien", "el", "sindicato", "respet\u00f3", "su",
            "decisi\u00f3n", "."],
 "text_c": ["club", "guipuzcoano", "no", "tiene", "opciones", "de", "acceder", "al", "mercado", "espa\u00f1ol", ",",
            "ya", "que", "a", "su", "juicio", "\"", "es", "imposible", "\"", "fichar", "jugadores", "de", "\"",
            "cierto", "nivel", "\"", "sindical", "dijo", "tras", "la", "desconvocatoria", "de", "la", "huelga",
            "que", "este", "preacuerdo", "\"", "puede", "ser", "perjudicial", "y", "regresivo", "para"],
 "span_a": [12, 39], "span_b": [2, 20], "span_c_a": [0, 27], "span_c_b": [27, 45]},
{"text_a": ["Todos", "estos", "temas", "se", "tratar\u00e1n", "en", "la", "Comisi\u00f3n", "Mixta", "RENFE-Junta",
            "de", "Comunidades", ",", "creada", "para", "negociar", "todos", "los", "problemas", "que", "existen",
            "al", "respecto", "en", "la", "Comunidad", "Aut\u00f3noma", "y", "de", "la", "cual", "formar\u00e1",
            "parte", "el", "Ayuntamiento", "de", "Alc\u00e1zar", "como", "observador", "."],
 "text_b": ["Este", "reconocido", "profesional", ",", "con", "estudios", "en", "Nueva", "York", ",", "Tokio",
            "y", "Buenos", "Aires", ",", "tiene", "una", "amplia", "experiencia", "en", "el", "dise\u00f1o",
            "de", "espacios", "multidisciplinares", "y", "con", "fines", "culturales", "y", "entre", "otros",
            "proyectos", "trabaja", "en", "la", "actualidad", "en", "el", "dise\u00f1o", "de", "un", "centro",
            "cultural", "para", "Filadelfia", ",", "que", "albergar\u00e1", "a", "la", "orquesta", "titular",
            "de", "esa", "ciudad", "."],
 "text_c": ["al", "respecto", "en", "la", "Comunidad", "Aut\u00f3noma", "y", "de", "la", "cual", "formar\u00e1",
            "parte", "el", "Ayuntamiento", "de", "Alc\u00e1zar", "como", "observador", ".", "experiencia", "en",
            "el", "dise\u00f1o", "de", "espacios", "multidisciplinares", "y", "con", "fines", "culturales", "y",
            "entre", "otros", "proyectos", "trabaja", "en", "la", "actualidad", "en", "el", "dise\u00f1o", "de",
            "un", "centro", "cultural", "para", "Filadelfia"],
 "span_a": [21, 40], "span_b": [18, 46], "span_c_a": [0, 19], "span_c_b": [19, 47]},
{"text_a": ["Posibilidad", "de", "alg\u00fan", "banco", "de", "niebla", "en", "el", "sureste", "."],
 "text_b": ["Temperaturas", "en", "ascenso", "ligero", "en", "el", "\u00e1rea", "del", "cant\u00e1brico",
            "oriental", ",", "La", "Rioja", ",", "Navarra", ",", "Arag\u00f3n", "y", "en", "Canarias", "y",
            "sin", "cambios", "en", "el", "resto", "."],
 "text_c": ["de", "niebla", "en", "el", "sureste", "en", "ascenso", "ligero", "en", "el", "\u00e1rea", "del",
            "cant\u00e1brico", "oriental", ",", "La", "Rioja"],
 "span_a": [4, 9], "span_b": [1, 13], "span_c_a": [0, 5], "span_c_b": [5, 17]}]

dataflow.update_with_jsons(examples)

def check_example(example):
  a_tokens = example.a_tokens
  b_tokens = example.b_tokens
  c_tokens = example.c_tokens
  span_a_selected_index = example.span_a_selected_index
  span_b_selected_index = example.span_b_selected_index
  span_c_a_selected_index = example.span_c_a_selected_index
  span_c_b_selected_index = example.span_c_b_selected_index
  assert((len(span_a_selected_index) + len(span_b_selected_index)) ==
         (len(span_c_a_selected_index) + len(span_c_b_selected_index)))
  assert(a_tokens[span_a_selected_index[0]] == c_tokens[span_c_a_selected_index[0]])
  assert (b_tokens[span_b_selected_index[0]] == c_tokens[span_c_b_selected_index[0]])

for mb in dataflow.get_minibatches(minibatch_size=2):
  mb.generate_input("cpu", True)
  for example in mb.examples:
    check_example(example)

print("Passed.")




import torch
from torch import nn


def dynamic_rnn(rnn, inputs, seq_lengths):
  """
  :param rnn: RNN instance
  :param inputs: FloatTensor, shape [batch, time, dim] if rnn.batch_first else [time, batch, dim]
  :param seq_lengths: LongTensor shape [batch]
  :param hidden_state: FloatTensor,
  :return: the result of rnn layer
  """
  batch_first = rnn.batch_first
  sorted_seq_lengths, indices = torch.sort(seq_lengths, descending=True)
  _, desorted_indices = torch.sort(indices, descending=False)
  if batch_first:
    inputs = inputs[indices]
  else:
    inputs = inputs[:, indices]
  packed_inputs = nn.utils.rnn.pack_padded_sequence(
    inputs,
    sorted_seq_lengths.cpu().numpy(),
    batch_first=batch_first)
  res, states = rnn(packed_inputs)
  padded_res, _ = nn.utils.rnn.pad_packed_sequence(res, batch_first=batch_first)
  if batch_first:
    desorted_res = padded_res[desorted_indices]
  else:
    desorted_res = padded_res[:, desorted_indices]
  if isinstance(states, tuple):
    states =  tuple(state[:, desorted_indices] for state in states)
  else:
    states = states[:, desorted_indices]
  return desorted_res, states


from dataclasses import dataclass
import json
from typing import Dict

@dataclass
class TrainerConfig:

  tokenizers: Dict

  def __post_init__(self):
    pass

  @classmethod
  def load_from_json(cls, config):
    return cls(
      tokenizers = config["tokenizers"]
    )

  @classmethod
  def load_from_namespace(cls, config):
    pass

  @classmethod
  def load_from_json_file(cls, config_path):
    return cls.load_from_json(json.load(open(config_path)))





import json
from types import SimpleNamespace

import relogic.utils.crash_on_ipy
from relogic.logickit.base.constants import DOCIR_TASK
from relogic.logickit.dataflow import TASK_TO_DATAFLOW_CLASS_MAP, DocPointwiseDataFlow
from relogic.logickit.tokenizer.tokenization import BertTokenizer

config = SimpleNamespace(
  **{
    "buckets": [(0, 15)],
    "max_seq_length": 450,
    "label_mapping_path": "data/preprocessed_data/binary_classification.json"
  })

tokenizers = {
  "BPE": BertTokenizer.from_pretrained("bert-base-multilingual-cased"),
}

dataflow: DocPointwiseDataFlow = TASK_TO_DATAFLOW_CLASS_MAP[DOCIR_TASK](
  task_name=DOCIR_TASK,
  config=config,
  tokenizers=tokenizers,
  label_mapping=json.load(open(config.label_mapping_path)))

examples = [{
  "text_a": "bbc world service staff cuts",
  "text_b":
  ["gossip day by day : bbc world service to cut five language services",
   "stoke city : begovic wilkinson shawcross wilson wilkinson walters whelan nzonzi kightly jerome crouch"],
  "label": "1"
}, {
  "text_a": "barbara walters chicken pox",
  "text_b": [
  "gossip day by day : bbc world service to cut five language services",
  "stoke city : begovic wilkinson shawcross wilson wilkinson walters whelan nzonzi kightly jerome crouch"],
  "label": "0"
}]

dataflow.update_with_jsons(examples)

for mb in dataflow.get_minibatches(minibatch_size=1):
  print(mb)

raise NotImplementedError("You can start to play with data")



import argparse
import json
from itertools import groupby

def main(answer_dump, answer_score, na_probs, threshold):
  # First group items based on the question
  answer_dump_group = dict(groupby(answer_dump.items(), key=lambda x: x[0].split('.')[0]))
  answer_score_group = dict(groupby(answer_score.items(), key=lambda x: x[0].split('.')[0]))
  na_probs_group = dict(groupby(na_probs.items(), key=lambda x: x[0].split('.')[0]))
  for qid in answer_dump_group:
    answers = answer_dump_group[qid]
    answer_scores = answer_score_group[qid]
    answer_na_probs = na_probs_group[qid]



if __name__ == "__main__":
  parser = argparse.ArgumentParser()
  parser.add_argument("--answer_dump_file")
  parser.add_argument("--answer_score_file")
  parser.add_argument("--na_prob_file")

  threshold = 1.0
  args = parser.parse_args()
  with open(args.answer_dump_file) as fin:
    answer_dump = json.load(fin)
  with open(args.answer_score_file) as fin:
    answer_score = json.load(fin)
  with open(args.na_prob_file) as fin:
    na_probs = json.load(fin)
  main(answer_dump, answer_score, na_probs, threshold)


"""Convert the fast_align
New York , 2-27 May 2005 ||| Nueva York , 2 a 27 de mayo de 2005
0-0 1-1 1-2 2-2 3-3 3-4 3-5 3-6 4-7 4-8 5-9

the expected output is
([0, 1, 3, 4, 5], [0, 1, 3, 7, 9])
The algorithm here is to use connect components.
"""
import argparse
from scipy.sparse import csr_matrix
from scipy.sparse.csgraph import connected_components
import numpy as np
import json

def main(alignment_file_path, parallel_file_path, selected_indices_file_path):
  fout = open(selected_indices_file_path, 'w')
  counter = 0
  with open(alignment_file_path) as fin_align, open(parallel_file_path) as fin_parallel:
    for alignment, sentence_pair in zip(fin_align, fin_parallel):
      text_a, text_b = sentence_pair.strip().split("|||")
      a_tokens, b_tokens = text_a.strip().split(), text_b.strip().split()
      length = len(a_tokens) + len(b_tokens)
      shape = (length, length)
      a_base_index, b_base_index = 0, len(a_tokens)
      text_a_index_array, text_b_index_array, indicator_array = [], [], []
      for pair in alignment.strip().split():
        index_a, index_b = pair.split('-')
        index_a = int(index_a) + a_base_index
        index_b = int(index_b) + b_base_index
        text_a_index_array.append(index_a)
        text_b_index_array.append(index_b)
        indicator_array.append(1)
      matrix = csr_matrix((np.array(indicator_array), (np.array(text_a_index_array), np.array(text_b_index_array))), shape=shape)
      n_components, labels = connected_components(csgraph=matrix, directed=False, return_labels=True)
      a_components, b_components = [], []
      for i in range(n_components):
        a_components.append([])
        b_components.append([])
      for i in range(len(a_tokens)):
        a_components[labels[a_base_index+i]].append(i)
      for i in range(len(b_tokens)):
        b_components[labels[b_base_index+i]].append(i)
      a_selected_index, b_selected_index = [], []
      for a_component, b_component in zip(a_components, b_components):
        if len(a_component) > 0 and len(b_component) > 0:
          a_selected_index.append(min(a_component))
          b_selected_index.append(min(b_component))
      fout.write(json.dumps([a_selected_index, b_selected_index]) + "\n")
      counter += 1
      if counter % 100000 == 0:
        print(counter)

if __name__ == "__main__":
  parser = argparse.ArgumentParser()
  parser.add_argument("--alignment_file_path", type=str)
  parser.add_argument("--parallel_file_path", type=str)
  parser.add_argument("--selected_indices_file_path", type=str)
  args = parser.parse_args()
  main(args.alignment_file_path, args.parallel_file_path, args.selected_indices_file_path)



from dataclasses import dataclass
import json
from typing import Dict
import logging
from types import SimpleNamespace
import os
import relogic

logger = logging.getLogger(__name__)

class Configs:
  def __init__(self):
    pass

class EncoderConfigs(Configs):
  def __init__(self, encoder_configs: Dict):
    super().__init__()
    self.fix_embedding = encoder_configs.get("fix_embedding", False)

    self.fix_layers = []
    fix_layers = encoder_configs.get("fix_layers", list([]))
    for layer_range in fix_layers:
      if '-' in layer_range:
        s, e = layer_range.split('-')
        for i in range(int(s), int(e)+1): # inclusive
          self.fix_layers.append(i)
      else:
        self.fix_layers.append(int(layer_range))

class AdversarialConfigs(Configs):
  def __init__(self, adversarial_configs: Dict):
    super().__init__()
    self.activate = len(adversarial_configs) > 0
    self.type = adversarial_configs.get("type")
    self.discriminator_type = adversarial_configs.get("discriminator_type")
    self.discriminator_lr = adversarial_configs.get("adversarial_lr")
    self.soft_label = adversarial_configs.get("soft_label", False)
    self.clip_upper = adversarial_configs.get("clip_upper", 0.01)
    self.clip_lower = adversarial_configs.get("clip_lower", -0.01)
    self.scale = adversarial_configs.get("scale", 0.01)
    self.hidden_size = adversarial_configs.get("hidden_size", 798)

class ParamConfigs(Configs):
  def __init__(self, param_configs: Dict):
    super().__init__()
    self.param_assignment = param_configs.get("param_assignment", None)
    if self.param_assignment is not None:
      logger.info("The following parameters will be replaced")
      logger.info(self.param_assignment)


class Configuration:
  def __init__(
        self,
        tokenizer_configs: Dict,
        module_configs: Dict,
        task_configs: Dict,
        encoder_configs: EncoderConfigs,
        adversarial_configs: AdversarialConfigs,
        param_configs):
    self.tokenizer_configs = tokenizer_configs
    self.module_configs = module_configs
    self.task_configs = task_configs
    self.encoder_configs = encoder_configs
    self.adversarial_configs = adversarial_configs
    self.param_configs = param_configs
    if self.adversarial_configs.activate:
      logger.info(self.adversarial_configs.__dict__)


  @property
  def module_names(self):
    return self.module_configs.keys()

  @property
  def task_names(self):
    return self.task_configs.keys()

  @classmethod
  def load_from_json(cls, config):
    return cls(
      tokenizer_configs = config["tokenizers"],
      module_configs = config.get("modules"),
      task_configs = config.get("tasks"),
      encoder_configs = EncoderConfigs(config.get("encoder", dict({}))),
      adversarial_configs = AdversarialConfigs(config.get("adversarial", dict({}))),
      param_configs = ParamConfigs(config.get("param", dict({})))
    )

  @classmethod
  def load_from_namespace(cls, config):
    pass

  @classmethod
  def load_from_json_file(cls, config_path):
    return cls.load_from_json(json.load(open(config_path)))

class Argument:
  def __init__(self):
    pass

  @classmethod
  def restore_from_model_path(cls, model_path, config_name="general_config.json", mode="deployment"):
    with open(os.path.join(model_path, config_name)) as f:
      restore_config = SimpleNamespace(**json.load(f))
    restore_config.mode = "deployment"
    restore_config.adversarial_training = False
    for task in restore_config.tasks:
      origin_label_path = restore_config.tasks[task]["label_mapping_path"]
      if origin_label_path != "none":
        label_mapping_path = origin_label_path.split("/")[-2:]
        absolute_label_mapping_path = os.path.join(os.path.dirname(relogic.__file__), "..", "data", *label_mapping_path)
        restore_config.tasks[task]["label_mapping_path"] = absolute_label_mapping_path
    restore_config.config_file = os.path.join(os.path.dirname(relogic.__file__), "..", restore_config.config_file)
    if restore_config.training_scheme_file:
      restore_config.training_scheme_file = os.path.join(os.path.dirname(relogic.__file__), "..", restore_config.training_scheme_file)
    return restore_config


from relogic.components.component import Component
from relogic.structures.structure import Structure
from typing import List

class EntityLinkingComponent(Component):
  """For given entity mention, this component is to 
  link the mention to a given knowledge graph.
  """
  
  def execute(self, inputs: List[Structure]):
    """All entity mentions in the inputs are given.
    Retrieve a candidate list regarding each mention.
    Then the global inference is operated to decide the top
      ranked item for each mention.
    """
    


import torch
import torch.nn as nn

class SpanGCNModule(nn.Module):
  """
  SpanGCN firstly extract span from text, and then label each span based
    on the learned representation of GCN
  """
  def __init__(self, config, task_name, boundary_n_classes=None, label_n_classes=None):
    super(SpanGCNModule, self).__init__()
    self.config = config
    self.task_name = task_name
    self.boundary_n_classes = boundary_n_classes
    self.label_n_classes = label_n_classes
    if boundary_n_classes:
      self.to_boundary_logits = nn.Linear(config.hidden_size, self.boundary_n_classes)
    if label_n_classes:
      self.to_label_logits = nn.Linear(config.hidden_size * 2, self.label_n_classes)
    if config.use_gcn:
      pass
    else:
      pass
    self.padding = nn.Parameter(torch.zeros(config.hidden_size), requires_grad=False)
    self.ones = nn.Parameter(torch.ones(1, 1), requires_grad=False)

  def forward(self,
              input, predicate_span=None,
              bio_hidden=None, span_candidates=None, extra_args=None, **kwargs):
    """
    Before this module, there is another module info aggregation
    :param input: Sentence Only, in batch
    :param predicate: Predicate Only, in batch
    :param bio_hidden: hidden vector for span prediction, can be None
    :param span_candidates: tuple, span_start, and span_end
    :param extra_args: strategy
    :return: labeling_logits

    Here we need to support three mode of inference
    1. Span is given
       In this mode, sequence labeling and independent span generation modes are supported.
       span_logits = None, span_candidates = batch of span
       Another problem here is how to aggregate span. We need to specify span level and aggregate method
       - For token level span
         - Pooling
         - Average
         - Attentive
         - Head Tail Attentive
       - For phrase level span
         - Non-Hierarchy
           - Pooling
           - Average
           - Attentive
           - Head Tail Attentive
         - Hierarchy
           - Use Token level aggregate
           - Use Non-Hierarchy to aggregate again
    2. Span is not given
       In this mode, dependent span generation mode is supported.
       span_logits = batch, span_candidates = None
       span candidates generate from span_logits. There are two logits need to be return
       This mode only have Phrase level span Aggregation

    After we have span representation, how to interact with predicate
    1. If the span itself is predicate aware, do we need to add predicate information again ?
       So the experiments is on surface form aware or not. You can refer the IBM relation paper
    2. If the span itself is not predicate aware, it will be trained faster.
       How to design a module to interact the argument and the predicate
       - Independent Classifier
       - Bilinear
       - GCN
    """
    if bio_hidden:
      bio_logits = self.to_span_logits(bio_hidden)
      assert "label_mapping" in extra_args, "label_mapping does not in extra_args"
      span_candidates = get_candidate_span(bio_logits, extra_args["label_mapping"])
    start_index, end_index = span_candidates
    # start_index, end_index = (batch, max_span_num)
    predicate_start_index, predicate_end_index = predicate_span
    # predicate_start_index, predicate_end_index = (batch)
    max_span_num = len(start_index[0])
    # input (batch, sentence, dim) -> (batch, max_span_num, sentence, dim)
    expanded_input = input.unsqueeze(1).repeat(1, max_span_num, 1, 1)
    start_index_ = start_index.view(-1)
    end_index_ = end_index.view(-1)

    span_hidden = select_span(expanded_input.view(-1, expanded_input.size(-2), expanded_input.size(-1)), start_index_, end_index_, self.padding)
    predicate_hidden = select_span(input, predicate_start_index, predicate_end_index, self.padding)

    span_repr = self.aggregate(span_hidden, end_index_-start_index_)
    predicate_repr = self.aggregate(predicate_hidden, predicate_end_index-predicate_start_index)
    # (batch, dim)
    concat = torch.cat([span_repr, predicate_repr.unsqueeze(1).repeat(1, max_span_num, 1).view(-1, predicate_repr.size(-1))], dim=-1)
    label_logits = self.to_label_logits(concat)

    return label_logits.view(input.size(0), max_span_num, self.label_n_classes)

  def aggregate(self, hidden, lengths):
    """
    Use average for now
    :param hidden: (batch, span_length, dim)
    :param lengths: (batch)
    :return:
    """
    return torch.sum(hidden, 1) / torch.max(
      self.ones.repeat(lengths.size(0), 1).float(), lengths.unsqueeze(1).float())

def select_span(input, start_index, end_index, padding):
  """
  Use for loop to select
  :param input:
  :param start_index:
  :param end_index:
  :param padding:
  :return:
  """
  padded_tensor = []

  max_span_size = torch.max(end_index - start_index)
  for idx, (start, end) in enumerate(zip(start_index, end_index)):
    padded_tensor.append(
      torch.cat(
        [torch.narrow(input[idx], 0, start, (end-start))] +
        ([padding.unsqueeze(0).repeat(max_span_size-(end-start), 1)] if max_span_size != (end-start)
        else []), dim=0))
    # list of (max_span_size, dim)
  return torch.stack(padded_tensor)

def get_candidate_span(bio_logits, label_mapping):
  """
  Use python for now. Will consider a C++ binding later.
  :param bio_logits: (batch_size, sentence_length, 3)
  :param label_mappings:
  :return:
  """
  preds_tags = bio_logits.argmax(-1).data.cpu().numpy()
  inv_label_mapping = {v: k for k, v in label_mapping.items()}
  batch_span_labels = []
  max_span_num = 0
  for sentence in bio_logits:
    # convert to understandable labels
    sentence_tags = [inv_label_mapping[i] for i in sentence]
    span_labels = []
    last = 'O'
    start = -1
    for i, tag in enumerate(sentence_tags):
      pos, _ = (None, 'O') if tag == 'O' else tag.split('-', 1)
      if (pos == 'S' or pos == 'B' or tag == 'O') and last != 'O':
        span_labels.append((start, i - 1, last.split('-')[-1]))
      if pos == 'B' or pos == 'S' or last == 'O':
        start = i
      last = tag
    if sentence_tags[-1] != 'O':
      span_labels.append((start, len(sentence_tags) - 1,
                          sentence_tags[-1].split('-', 1)[-1]))
    max_span_num = max(len(span_labels), max_span_num)
    batch_span_labels.append(span_labels)
  batch_start_index = []
  batch_end_index = []
  for span_labels in batch_span_labels:
    start_index = []
    end_index = []
    for span_label in span_labels:
      start_index.append(span_label[0])
      end_index.append(span_label[1])
    start_index += (max_span_num - len(start_index)) * [0]
    end_index += (max_span_num - len(end_index)) * [0]
    # Just a placeholder, for loss computation, it will be ignored.
    batch_start_index.append(start_index)
    batch_end_index.append(end_index)
  start_ids = torch.tensor(batch_start_index, dtype=torch.long).to(bio_logits.device)
  end_ids = torch.tensor(batch_end_index, dtype=torch.long).to(bio_logits.device)
  return (start_ids, end_ids)


